# Copyright (C) 2012 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Backup manager manages volume backups.

Volume Backups are full copies of persistent volumes stored in a backup
store e.g. an object store or any other backup store if and when support is
added. They are usable without the original object being available. A
volume backup can be restored to the original volume it was created from or
any other available volume with a minimum size of the original volume.
Volume backups can be created, restored, deleted and listed.

**Related Flags**

:backup_manager:  The module name of a class derived from
                          :class:`manager.Manager` (default:
                          :class:`cinder.backup.manager.Manager`).

"""

import contextlib
import os

from castellan import key_manager
from eventlet import tpool
from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_service import loopingcall
from oslo_service import periodic_task
from oslo_utils import excutils
from oslo_utils import importutils
from oslo_utils import timeutils

from cinder.backup import rpcapi as backup_rpcapi
from cinder import context
from cinder import exception
from cinder.i18n import _
from cinder.keymgr import migration as key_migration
from cinder import manager
from cinder.message import api as message_api
from cinder.message import message_field
from cinder import objects
from cinder.objects import fields
from cinder import quota
from cinder import utils
from cinder.volume import rpcapi as volume_rpcapi
from cinder.volume import volume_utils

LOG = logging.getLogger(__name__)

backup_manager_opts = [
    cfg.StrOpt('backup_driver',
               default='cinder.backup.drivers.swift.SwiftBackupDriver',
               help='Driver to use for backups.',),
    cfg.IntOpt('backup_driver_init_check_interval',
               default=60,
               min=5,
               help='Time in seconds between checks to see if the backup '
                    'driver has been successfully initialized, any time '
                    'the driver is restarted.'),
    cfg.IntOpt('backup_driver_stats_polling_interval',
               default=60,
               min=10,
               deprecated_name='backup_driver_status_check_interval',
               help='Time in seconds between checks of the backup driver '
                    'status.  If does not report as working, it is '
                    'restarted.'),
    cfg.BoolOpt('backup_service_inithost_offload',
                default=True,
                help='Offload pending backup delete during '
                     'backup service startup. If false, the backup service '
                     'will remain down until all pending backups are '
                     'deleted.',),
    cfg.IntOpt('backup_native_threads_pool_size',
               default=60,
               min=20,
               help='Size of the native threads pool for the backups.  '
                    'Most backup drivers rely heavily on this, it can be '
                    'decreased for specific drivers that don\'t.'),
]

CONF = cfg.CONF
CONF.register_opts(backup_manager_opts)
CONF.import_opt('use_multipath_for_image_xfer', 'cinder.volume.driver')
CONF.import_opt('num_volume_device_scan_tries', 'cinder.volume.driver')
QUOTAS = quota.QUOTAS
MAPPING = {
    # Module name "google" conflicts with google library namespace inside the
    # driver when it imports google.auth
    'cinder.backup.drivers.google.GoogleBackupDriver':
    'cinder.backup.drivers.gcs.GoogleBackupDriver',
}
SERVICE_PGRP = '' if os.name == 'nt' else os.getpgrp()


# TODO(geguileo): Once Eventlet issue #432 gets fixed we can just tpool.execute
# the whole call to the driver's backup and restore methods instead of proxy
# wrapping the device_file and having the drivers also proxy wrap their
# writes/reads and the compression/decompression calls.
# (https://github.com/eventlet/eventlet/issues/432)

class BackupManager(manager.SchedulerDependentManager):
    """Manages backup of block storage devices."""

    RPC_API_VERSION = backup_rpcapi.BackupAPI.RPC_API_VERSION

    target = messaging.Target(version=RPC_API_VERSION)

    def __init__(self, *args, **kwargs):
        self.az = CONF.storage_availability_zone
        self.backup_rpcapi = backup_rpcapi.BackupAPI()
        self.volume_rpcapi = volume_rpcapi.VolumeAPI()
        super(BackupManager, self).__init__(*args, **kwargs)
        self.is_initialized = False
        self._set_tpool_size(CONF.backup_native_threads_pool_size)
        self._process_number = kwargs.get('process_number', 1)
        self._semaphore = kwargs.get('semaphore', contextlib.suppress())
        self.driver_name = CONF.backup_driver
        if self.driver_name in MAPPING:
            new_name = MAPPING[self.driver_name]
            LOG.warning('Backup driver path %s is deprecated, update your '
                        'configuration to the new path %s',
                        self.driver_name, new_name)
            self.driver_name = new_name
        self.service = importutils.import_class(self.driver_name)
        self.message_api = message_api.API()

    def init_host(self, **kwargs):
        """Run initialization needed for a standalone service."""
        ctxt = context.get_admin_context()
        self.setup_backup_backend(ctxt)

        try:
            self._cleanup_incomplete_backup_operations(ctxt)
        except Exception:
            # Don't block startup of the backup service.
            LOG.exception("Problem cleaning incomplete backup operations.")

        # Migrate any ConfKeyManager keys based on fixed_key to the currently
        # configured key manager.
        backups = objects.BackupList.get_all_by_host(ctxt, self.host)
        self._add_to_threadpool(key_migration.migrate_fixed_key,
                                backups=backups)
        self.publish_service_capabilities(ctxt)

    def _setup_backup_driver(self, ctxt):
        backup_service = self.service(context=ctxt)
        backup_service.check_for_setup_error()
        self.is_initialized = True
        raise loopingcall.LoopingCallDone()

    def setup_backup_backend(self, ctxt):
        try:
            init_loop = loopingcall.FixedIntervalLoopingCall(
                self._setup_backup_driver, ctxt)
            init_loop.start(interval=CONF.backup_driver_init_check_interval)
        except loopingcall.LoopingCallDone:
            LOG.info("Backup driver was successfully initialized.")
        except Exception:
            LOG.exception("Failed to initialize driver.",
                          resource={'type': 'driver',
                                    'id': self.__class__.__name__})

    def reset(self):
        super(BackupManager, self).reset()
        self.backup_rpcapi = backup_rpcapi.BackupAPI()
        self.volume_rpcapi = volume_rpcapi.VolumeAPI()

    @utils.synchronized('cleanup_incomplete_backups_%s' % SERVICE_PGRP,
                        external=True, delay=0.1)
    def _cleanup_incomplete_backup_operations(self, ctxt):
        # Only the first launched process should do the cleanup, the others
        # have waited on the lock for the first one to finish the cleanup and
        # can now continue with the start process.
        if self._process_number != 1:
            LOG.debug("Process #%s %sskips cleanup.",
                      self._process_number,
                      '(pgid=%s) ' % SERVICE_PGRP if SERVICE_PGRP else '')
            return

        LOG.info("Cleaning up incomplete backup operations.")

        # TODO(smulcahy) implement full resume of backup and restore
        # operations on restart (rather than simply resetting)
        backups = objects.BackupList.get_all_by_host(ctxt, self.host)
        for backup in backups:
            try:
                self._cleanup_one_backup(ctxt, backup)
            except Exception:
                LOG.exception("Problem cleaning up backup %(bkup)s.",
                              {'bkup': backup['id']})
            try:
                self._cleanup_temp_volumes_snapshots_for_one_backup(ctxt,
                                                                    backup)
            except Exception:
                LOG.exception("Problem cleaning temp volumes and "
                              "snapshots for backup %(bkup)s.",
                              {'bkup': backup['id']})

    def _cleanup_one_volume(self, ctxt, volume):
        if volume['status'] == 'backing-up':
            self._detach_all_attachments(ctxt, volume)
            LOG.info('Resetting volume %(vol_id)s to previous '
                     'status %(status)s (was backing-up).',
                     {'vol_id': volume['id'],
                      'status': volume['previous_status']})
            self.db.volume_update(ctxt, volume['id'],
                                  {'status': volume['previous_status']})
        elif volume['status'] == 'restoring-backup':
            self._detach_all_attachments(ctxt, volume)
            LOG.info('Setting volume %s to error_restoring '
                     '(was restoring-backup).', volume['id'])
            self.db.volume_update(ctxt, volume['id'],
                                  {'status': 'error_restoring'})

    def _cleanup_one_backup(self, ctxt, backup):
        if backup['status'] == fields.BackupStatus.CREATING:
            LOG.info('Resetting backup %s to error (was creating).',
                     backup['id'])

            volume = objects.Volume.get_by_id(ctxt, backup.volume_id)
            self._cleanup_one_volume(ctxt, volume)

            err = 'incomplete backup reset on manager restart'
            volume_utils.update_backup_error(backup, err)
        elif backup['status'] == fields.BackupStatus.RESTORING:
            LOG.info('Resetting backup %s to '
                     'available (was restoring).',
                     backup['id'])
            volume = objects.Volume.get_by_id(ctxt, backup.restore_volume_id)
            self._cleanup_one_volume(ctxt, volume)

            backup.status = fields.BackupStatus.AVAILABLE
            backup.save()
        elif backup['status'] == fields.BackupStatus.DELETING:
            # Don't resume deleting the backup of an encrypted volume. The
            # admin context won't be sufficient to delete the backup's copy
            # of the encryption key ID (a real user context is required).
            if backup.encryption_key_id is None:
                LOG.info('Resuming delete on backup: %s.', backup.id)
                if CONF.backup_service_inithost_offload:
                    # Offload all the pending backup delete operations to the
                    # threadpool to prevent the main backup service thread
                    # from being blocked.
                    self._add_to_threadpool(self.delete_backup, ctxt, backup)
                else:
                    # Delete backups sequentially
                    self.delete_backup(ctxt, backup)
            else:
                LOG.info('Unable to resume deleting backup of an encrypted '
                         'volume, resetting backup %s to error_deleting '
                         '(was deleting).',
                         backup.id)
                backup.status = fields.BackupStatus.ERROR_DELETING
                backup.save()

    def _detach_all_attachments(self, ctxt, volume):
        attachments = volume['volume_attachment'] or []
        for attachment in attachments:
            if (attachment['attached_host'] == self.host and
                    attachment['instance_uuid'] is None):
                try:
                    rpcapi = self.volume_rpcapi
                    rpcapi.detach_volume(ctxt, volume, attachment['id'])
                except Exception:
                    LOG.exception("Detach attachment %(attach_id)s failed.",
                                  {'attach_id': attachment['id']},
                                  resource=volume)

    def _delete_temp_volume(self, ctxt, backup):
        try:
            temp_volume = objects.Volume.get_by_id(
                ctxt, backup.temp_volume_id)
            self.volume_rpcapi.delete_volume(ctxt, temp_volume)
        except exception.VolumeNotFound:
            LOG.debug("Could not find temp volume %(vol)s to clean up "
                      "for backup %(backup)s.",
                      {'vol': backup.temp_volume_id,
                       'backup': backup.id})
        backup.temp_volume_id = None
        backup.save()

    def _delete_temp_snapshot(self, ctxt, backup):
        try:
            temp_snapshot = objects.Snapshot.get_by_id(
                ctxt, backup.temp_snapshot_id)
            # We may want to consider routing those calls through the
            # cinder API.
            temp_snapshot.status = fields.SnapshotStatus.DELETING
            temp_snapshot.save()
            self.volume_rpcapi.delete_snapshot(ctxt, temp_snapshot)
        except exception.SnapshotNotFound:
            LOG.debug("Could not find temp snapshot %(snap)s to clean "
                      "up for backup %(backup)s.",
                      {'snap': backup.temp_snapshot_id,
                       'backup': backup.id})
        backup.temp_snapshot_id = None
        backup.save()

    def _cleanup_temp_volumes_snapshots_for_one_backup(self, ctxt, backup):
        # NOTE(xyang): If the service crashes or gets restarted during the
        # backup operation, there could be temporary volumes or snapshots
        # that are not deleted. Make sure any temporary volumes or snapshots
        # create by the backup job are deleted when service is started.
        if (backup.temp_volume_id
                and backup.status == fields.BackupStatus.ERROR):
            self._delete_temp_volume(ctxt, backup)

        if (backup.temp_snapshot_id
                and backup.status == fields.BackupStatus.ERROR):
            self._delete_temp_snapshot(ctxt, backup)

    def _cleanup_temp_volumes_snapshots_when_backup_created(
            self, ctxt, backup):
        # Delete temp volumes or snapshots when backup creation is completed.
        if backup.temp_volume_id:
            self._delete_temp_volume(ctxt, backup)

        if backup.temp_snapshot_id:
            self._delete_temp_snapshot(ctxt, backup)

    @utils.limit_operations
    def create_backup(self, context, backup):
        """Create volume backups using configured backup service."""
        volume_id = backup.volume_id
        snapshot_id = backup.snapshot_id
        volume = objects.Volume.get_by_id(context, volume_id)
        snapshot = objects.Snapshot.get_by_id(
            context, snapshot_id) if snapshot_id else None
        previous_status = volume.get('previous_status', None)
        context.message_resource_id = backup.id
        context.message_resource_type = message_field.Resource.VOLUME_BACKUP
        context.message_action = message_field.Action.BACKUP_CREATE
        if snapshot_id:
            log_message = ('Create backup started, backup: %(backup_id)s '
                           'volume: %(volume_id)s snapshot: %(snapshot_id)s.'
                           % {'backup_id': backup.id,
                              'volume_id': volume_id,
                              'snapshot_id': snapshot_id})
        else:
            log_message = ('Create backup started, backup: %(backup_id)s '
                           'volume: %(volume_id)s.'
                           % {'backup_id': backup.id,
                              'volume_id': volume_id})
        LOG.info(log_message)

        self._notify_about_backup_usage(context, backup, "create.start")

        expected_status = "backing-up"
        if snapshot_id:
            actual_status = snapshot['status']
            if actual_status != expected_status:
                err = _('Create backup aborted, expected snapshot status '
                        '%(expected_status)s but got %(actual_status)s.') % {
                    'expected_status': expected_status,
                    'actual_status': actual_status,
                }
                volume_utils.update_backup_error(backup, err)
                raise exception.InvalidSnapshot(reason=err)
        else:
            actual_status = volume['status']
            if actual_status != expected_status:
                err = _('Create backup aborted, expected volume status '
                        '%(expected_status)s but got %(actual_status)s.') % {
                    'expected_status': expected_status,
                    'actual_status': actual_status,
                }
                volume_utils.update_backup_error(backup, err)
                raise exception.InvalidVolume(reason=err)

        expected_status = fields.BackupStatus.CREATING
        actual_status = backup.status
        if actual_status != expected_status:
            err = _('Create backup aborted, expected backup status '
                    '%(expected_status)s but got %(actual_status)s.') % {
                'expected_status': expected_status,
                'actual_status': actual_status,
            }
            volume_utils.update_backup_error(backup, err)
            self.message_api.create_from_request_context(
                context,
                detail=message_field.Detail.BACKUP_INVALID_STATE)
            raise exception.InvalidBackup(reason=err)

        try:
            if not self.is_working():
                err = _('Create backup aborted due to backup service is down.')
                volume_utils.update_backup_error(backup, err)
                self.message_api.create_from_request_context(
                    context,
                    detail=message_field.Detail.BACKUP_SERVICE_DOWN)
                raise exception.InvalidBackup(reason=err)

            backup.service = self.driver_name
            backup.save()

            # Start backup, then continue_backup, then finish_backup
            self._start_backup(context, backup, volume)
        except Exception as err:
            with excutils.save_and_reraise_exception():
                if snapshot_id:
                    snapshot.status = fields.SnapshotStatus.AVAILABLE
                    snapshot.save()
                else:
                    self.db.volume_update(
                        context, volume_id,
                        {'status': previous_status,
                         'previous_status': 'error_backing-up'})
                volume_utils.update_backup_error(backup, str(err))

    def _start_backup(self, context, backup, volume):
        """This starts the backup process.

           First we have to get the backup device from the volume manager.
           This can take a long time to complete.  Once the volume manager
           is done creating/getting the backup device, then we get a callback
           to complete the process of backing up the volume.

        """
        # Save a copy of the encryption key ID in case the volume is deleted.
        if (volume.encryption_key_id is not None and
                backup.encryption_key_id is None):
            backup.encryption_key_id = volume_utils.clone_encryption_key(
                context,
                key_manager.API(CONF),
                volume.encryption_key_id)
            backup.save()

        # This is an async call to the volume manager.  We will get a
        # callback from the volume manager to continue once it's done.
        LOG.info("Call Volume Manager to get_backup_device for %s", backup)
        self.volume_rpcapi.get_backup_device(context, backup, volume)

    def continue_backup(self, context, backup, backup_device):
        """This is the callback from the volume manager to continue."""
        message_created = False
        volume_id = backup.volume_id
        volume = objects.Volume.get_by_id(context, volume_id)
        snapshot_id = backup.snapshot_id
        snapshot = objects.Snapshot.get_by_id(
            context, snapshot_id) if snapshot_id else None
        previous_status = volume.get('previous_status', None)

        backup_service = self.service(context)
        properties = volume_utils.brick_get_connector_properties(
            CONF.use_multipath_for_image_xfer, enforce_multipath=False)

        updates = {}
        try:
            try:
                attach_info = self._attach_device(context,
                                                  backup_device.device_obj,
                                                  properties,
                                                  backup_device.is_snapshot)
            except Exception:
                with excutils.save_and_reraise_exception():
                    if not message_created:
                        message_created = True
                        self.message_api.create_from_request_context(
                            context,
                            detail=message_field.Detail.ATTACH_ERROR)

            try:
                device_path = attach_info['device']['path']
                if (isinstance(device_path, str) and
                        not os.path.isdir(device_path)):
                    if backup_device.secure_enabled:
                        with open(device_path, 'rb') as device_file:
                            updates = backup_service.backup(
                                backup, tpool.Proxy(device_file))
                    else:
                        with utils.temporary_chown(device_path):
                            with open(device_path, 'rb') as device_file:
                                updates = backup_service.backup(
                                    backup, tpool.Proxy(device_file))
                # device_path is already file-like so no need to open it
                else:
                    updates = backup_service.backup(backup,
                                                    tpool.Proxy(device_path))
            except Exception:
                with excutils.save_and_reraise_exception():
                    if not message_created:
                        message_created = True
                        self.message_api.create_from_request_context(
                            context,
                            detail=
                            message_field.Detail.BACKUP_CREATE_DRIVER_ERROR)
            finally:
                try:
                    self._detach_device(context, attach_info,
                                        backup_device.device_obj, properties,
                                        backup_device.is_snapshot, force=True,
                                        ignore_errors=True)
                except Exception:
                    with excutils.save_and_reraise_exception():
                        if not message_created:
                            message_created = True
                            self.message_api.create_from_request_context(
                                context,
                                detail=
                                message_field.Detail.DETACH_ERROR)
        except Exception as err:
            with excutils.save_and_reraise_exception():
                if snapshot_id:
                    snapshot.status = fields.SnapshotStatus.AVAILABLE
                    snapshot.save()
                else:
                    self.db.volume_update(
                        context, volume_id,
                        {'status': previous_status,
                         'previous_status': 'error_backing-up'})
                volume_utils.update_backup_error(backup, str(err))
        finally:
            with backup.as_read_deleted():
                backup.refresh()
            try:
                self._cleanup_temp_volumes_snapshots_when_backup_created(
                    context, backup)
            except Exception:
                with excutils.save_and_reraise_exception():
                    if not message_created:
                        self.message_api.create_from_request_context(
                            context,
                            detail=
                            message_field.Detail.BACKUP_CREATE_CLEANUP_ERROR)

        self._finish_backup(context, backup, volume, updates)

    def _finish_backup(self, context, backup, volume, updates):
        volume_id = backup.volume_id
        snapshot_id = backup.snapshot_id
        previous_status = volume.get('previous_status', None)

        # Restore the original status.
        if snapshot_id:
            self.db.snapshot_update(
                context, snapshot_id,
                {'status': fields.SnapshotStatus.AVAILABLE})
        else:
            self.db.volume_update(context, volume_id,
                                  {'status': previous_status,
                                   'previous_status': 'backing-up'})

        # _run_backup method above updated the status for the backup, so it
        # will reflect latest status, even if it is deleted
        completion_msg = 'finished'
        if backup.status in (fields.BackupStatus.DELETING,
                             fields.BackupStatus.DELETED):
            completion_msg = 'aborted'
        else:
            backup.status = fields.BackupStatus.AVAILABLE
            backup.size = volume['size']

            if updates:
                backup.update(updates)
            backup.save()

            # Handle the num_dependent_backups of parent backup when child
            # backup has created successfully.
            if backup.parent_id:
                parent_backup = objects.Backup.get_by_id(context,
                                                         backup.parent_id)
                parent_backup.num_dependent_backups += 1
                parent_backup.save()
        LOG.info('Create backup %s. backup: %s.', completion_msg, backup.id)
        self._notify_about_backup_usage(context, backup, "create.end")

    def _is_our_backup(self, backup):
        # Accept strings and Service OVO
        if not isinstance(backup, str):
            backup = backup.service

        if not backup:
            return True

        # TODO(tommylikehu): We upgraded the 'driver_name' from module
        # to class name, so we use 'in' here to match two namings,
        # this can be replaced with equal sign during next
        # release (Rocky).
        if self.driver_name.startswith(backup):
            return True

        # We support renaming of drivers, so check old names as well
        for key, value in MAPPING.items():
            if key.startswith(backup) and self.driver_name.startswith(value):
                return True

        return False

    @utils.limit_operations
    def restore_backup(self, context, backup, volume_id):
        """Restore volume backups from configured backup service."""
        context.message_resource_id = backup.id
        context.message_resource_type = message_field.Resource.VOLUME_BACKUP
        context.message_action = message_field.Action.BACKUP_RESTORE
        LOG.info('Restore backup started, backup: %(backup_id)s '
                 'volume: %(volume_id)s.',
                 {'backup_id': backup.id, 'volume_id': volume_id})

        volume = objects.Volume.get_by_id(context, volume_id)
        self._notify_about_backup_usage(context, backup, "restore.start")

        expected_status = [fields.VolumeStatus.RESTORING_BACKUP,
                           fields.VolumeStatus.CREATING]
        volume_previous_status = volume['status']
        if volume_previous_status not in expected_status:
            err = (_('Restore backup aborted, expected volume status '
                     '%(expected_status)s but got %(actual_status)s.') %
                   {'expected_status': ','.join(expected_status),
                    'actual_status': volume_previous_status})
            backup.status = fields.BackupStatus.AVAILABLE
            backup.save()
            self.db.volume_update(
                context, volume_id,
                {'status':
                 (fields.VolumeStatus.ERROR if
                  volume_previous_status == fields.VolumeStatus.CREATING else
                  fields.VolumeStatus.ERROR_RESTORING)})
            self.message_api.create(
                context,
                action=message_field.Action.BACKUP_RESTORE,
                resource_type=message_field.Resource.VOLUME_BACKUP,
                resource_uuid=volume.id,
                detail=message_field.Detail.VOLUME_INVALID_STATE)
            raise exception.InvalidVolume(reason=err)

        expected_status = fields.BackupStatus.RESTORING
        actual_status = backup['status']
        if actual_status != expected_status:
            err = (_('Restore backup aborted: expected backup status '
                     '%(expected_status)s but got %(actual_status)s.') %
                   {'expected_status': expected_status,
                    'actual_status': actual_status})
            volume_utils.update_backup_error(backup, err)
            self.db.volume_update(context, volume_id,
                                  {'status': fields.VolumeStatus.ERROR})
            self.message_api.create_from_request_context(
                context,
                detail=message_field.Detail.BACKUP_INVALID_STATE)
            raise exception.InvalidBackup(reason=err)

        if volume['size'] > backup['size']:
            LOG.info('Volume: %(vol_id)s, size: %(vol_size)d is '
                     'larger than backup: %(backup_id)s, '
                     'size: %(backup_size)d, continuing with restore.',
                     {'vol_id': volume['id'],
                      'vol_size': volume['size'],
                      'backup_id': backup['id'],
                      'backup_size': backup['size']})

        if not self._is_our_backup(backup):
            err = _('Restore backup aborted, the backup service currently'
                    ' configured [%(configured_service)s] is not the'
                    ' backup service that was used to create this'
                    ' backup [%(backup_service)s].') % {
                'configured_service': self.driver_name,
                'backup_service': backup.service,
            }
            backup.status = fields.BackupStatus.AVAILABLE
            backup.save()
            self.db.volume_update(context, volume_id,
                                  {'status': fields.VolumeStatus.ERROR})
            raise exception.InvalidBackup(reason=err)

        canceled = False
        try:
            self._run_restore(context, backup, volume)
        except exception.BackupRestoreCancel:
            canceled = True
        except Exception:
            with excutils.save_and_reraise_exception():
                self.db.volume_update(
                    context, volume_id,
                    {'status': (fields.VolumeStatus.ERROR if
                                actual_status == fields.VolumeStatus.CREATING
                                else fields.VolumeStatus.ERROR_RESTORING)})
                backup.status = fields.BackupStatus.AVAILABLE
                backup.save()

        if canceled:
            volume.status = fields.VolumeStatus.ERROR
        else:
            volume.status = fields.VolumeStatus.AVAILABLE
            # NOTE(tommylikehu): If previous status is 'creating', this is
            # just a new created volume and we need update the 'launched_at'
            # attribute as well.
            if volume_previous_status == fields.VolumeStatus.CREATING:
                volume['launched_at'] = timeutils.utcnow()
        old_src_backup_id = self.db.volume_metadata_get(
            context, volume_id).get("src_backup_id", None)
        if backup.volume_id != volume.id or (
                old_src_backup_id and old_src_backup_id != backup.id):
            self.db.volume_metadata_update(
                context,
                volume.id,
                {'src_backup_id': backup.id},
                False)

        volume.save()
        backup.status = fields.BackupStatus.AVAILABLE
        backup.save()
        LOG.info('%(result)s restoring backup %(backup_id)s to volume '
                 '%(volume_id)s.',
                 {'result': 'Canceled' if canceled else 'Finished',
                  'backup_id': backup.id,
                  'volume_id': volume_id})
        self._notify_about_backup_usage(context, backup, "restore.end")

    def _run_restore(self, context, backup, volume):
        message_created = False
        orig_key_id = volume.encryption_key_id
        backup_service = self.service(context)

        properties = volume_utils.brick_get_connector_properties(
            CONF.use_multipath_for_image_xfer, enforce_multipath=False)
        secure_enabled = (
            self.volume_rpcapi.secure_file_operations_enabled(context,
                                                              volume))
        try:
            attach_info = self._attach_device(context, volume, properties)
        except Exception:
            self.message_api.create_from_request_context(
                context,
                detail=message_field.Detail.ATTACH_ERROR)
            raise

        # NOTE(geguileo): Not all I/O disk operations properly do greenthread
        # context switching and may end up blocking the greenthread, so we go
        # with native threads proxy-wrapping the device file object.
        try:
            device_path = attach_info['device']['path']
            open_mode = 'rb+' if os.name == 'nt' else 'wb'
            if (isinstance(device_path, str) and
                    not os.path.isdir(device_path)):
                if secure_enabled:
                    with open(device_path, open_mode) as device_file:
                        backup_service.restore(backup, volume.id,
                                               tpool.Proxy(device_file))
                else:
                    with utils.temporary_chown(device_path):
                        with open(device_path, open_mode) as device_file:
                            backup_service.restore(backup, volume.id,
                                                   tpool.Proxy(device_file))
            # device_path is already file-like so no need to open it
            else:
                backup_service.restore(backup, volume.id,
                                       tpool.Proxy(device_path))
        except exception.BackupRestoreCancel:
            raise
        except Exception:
            LOG.exception('Restoring backup %(backup_id)s to volume '
                          '%(volume_id)s failed.', {'backup_id': backup.id,
                                                    'volume_id': volume.id})
            # We set message_create to True before creating the
            # message because if the message create call fails
            # and is catched by the base/outer exception handler
            # then we will end up storing a wrong message
            message_created = True
            self.message_api.create_from_request_context(
                context,
                detail=message_field.Detail.BACKUP_RESTORE_ERROR)
            raise
        finally:
            try:
                self._detach_device(context, attach_info, volume, properties,
                                    force=True)
            except Exception:
                if not message_created:
                    self.message_api.create_from_request_context(
                        context,
                        detail=message_field.Detail.DETACH_ERROR)
                raise

        # Regardless of whether the restore was successful, do some
        # housekeeping to ensure the restored volume's encryption key ID is
        # unique, and any previous key ID is deleted. Start by fetching fresh
        # info on the restored volume.
        restored_volume = objects.Volume.get_by_id(context, volume.id)
        restored_key_id = restored_volume.encryption_key_id
        if restored_key_id != orig_key_id:
            LOG.info('Updating encryption key ID for volume %(volume_id)s '
                     'from backup %(backup_id)s.',
                     {'volume_id': volume.id, 'backup_id': backup.id})

            key_mgr = key_manager.API(CONF)
            if orig_key_id is not None:
                LOG.debug('Deleting original volume encryption key ID.')
                volume_utils.delete_encryption_key(context,
                                                   key_mgr,
                                                   orig_key_id)

            if backup.encryption_key_id is None:
                # This backup predates the current code that stores the cloned
                # key ID in the backup database. Fortunately, the key ID
                # restored from the backup data _is_ a clone of the original
                # volume's key ID, so grab it.
                LOG.debug('Gleaning backup encryption key ID from metadata.')
                backup.encryption_key_id = restored_key_id
                backup.save()

            # Clone the key ID again to ensure every restored volume has
            # a unique key ID. The volume's key ID should not be the same
            # as the backup.encryption_key_id (the copy made when the backup
            # was first created).
            new_key_id = volume_utils.clone_encryption_key(
                context,
                key_mgr,
                backup.encryption_key_id)
            restored_volume.encryption_key_id = new_key_id
            restored_volume.save()
        else:
            LOG.debug('Encryption key ID for volume %(volume_id)s already '
                      'matches encryption key ID in backup %(backup_id)s.',
                      {'volume_id': volume.id, 'backup_id': backup.id})

    def delete_backup(self, context, backup):
        """Delete volume backup from configured backup service."""
        LOG.info('Delete backup started, backup: %s.', backup.id)

        self._notify_about_backup_usage(context, backup, "delete.start")

        context.message_resource_id = backup.id
        context.message_resource_type = message_field.Resource.VOLUME_BACKUP
        context.message_action = message_field.Action.BACKUP_DELETE
        expected_status = fields.BackupStatus.DELETING
        actual_status = backup.status
        if actual_status != expected_status:
            err = _('Delete_backup aborted, expected backup status '
                    '%(expected_status)s but got %(actual_status)s.') \
                % {'expected_status': expected_status,
                   'actual_status': actual_status}
            volume_utils.update_backup_error(backup, err)
            self.message_api.create_from_request_context(
                context,
                detail=message_field.Detail.BACKUP_INVALID_STATE)
            raise exception.InvalidBackup(reason=err)

        if backup.service and not self.is_working():
            err = _('Delete backup is aborted due to backup service is down.')
            status = fields.BackupStatus.ERROR_DELETING
            volume_utils.update_backup_error(backup, err, status)
            self.message_api.create_from_request_context(
                context,
                detail=message_field.Detail.BACKUP_SERVICE_DOWN)
            raise exception.InvalidBackup(reason=err)

        if not self._is_our_backup(backup):
            err = _('Delete backup aborted, the backup service currently'
                    ' configured [%(configured_service)s] is not the'
                    ' backup service that was used to create this'
                    ' backup [%(backup_service)s].')\
                % {'configured_service': self.driver_name,
                   'backup_service': backup.service}
            volume_utils.update_backup_error(backup, err)
            raise exception.InvalidBackup(reason=err)

        if backup.service:
            try:
                backup_service = self.service(context)
                backup_service.delete_backup(backup)
            except Exception as err:
                with excutils.save_and_reraise_exception():
                    volume_utils.update_backup_error(backup, str(err))
                    self.message_api.create_from_request_context(
                        context,
                        detail=message_field.Detail.BACKUP_DELETE_DRIVER_ERROR)

        # Get reservations
        try:
            reserve_opts = {
                'backups': -1,
                'backup_gigabytes': -backup.size,
            }
            reservations = QUOTAS.reserve(context,
                                          project_id=backup.project_id,
                                          **reserve_opts)
        except Exception:
            reservations = None
            LOG.exception("Failed to update usages deleting backup")

        if backup.encryption_key_id is not None:
            volume_utils.delete_encryption_key(context,
                                               key_manager.API(CONF),
                                               backup.encryption_key_id)
            backup.encryption_key_id = None
            backup.save()

        backup.destroy()
        # If this backup is incremental backup, handle the
        # num_dependent_backups of parent backup
        if backup.parent_id:
            parent_backup = objects.Backup.get_by_id(context,
                                                     backup.parent_id)
            if parent_backup.has_dependent_backups:
                parent_backup.num_dependent_backups -= 1
                parent_backup.save()
        # Commit the reservations
        if reservations:
            QUOTAS.commit(context, reservations,
                          project_id=backup.project_id)

        LOG.info('Delete backup finished, backup %s deleted.', backup.id)
        self._notify_about_backup_usage(context, backup, "delete.end")

    def _notify_about_backup_usage(self,
                                   context,
                                   backup,
                                   event_suffix,
                                   extra_usage_info=None):
        volume_utils.notify_about_backup_usage(
            context, backup, event_suffix,
            extra_usage_info=extra_usage_info,
            host=self.host)

    def export_record(self, context, backup):
        """Export all volume backup metadata details to allow clean import.

        Export backup metadata so it could be re-imported into the database
        without any prerequisite in the backup database.

        :param context: running context
        :param backup: backup object to export
        :returns: backup_record - a description of how to import the backup
        :returns: contains 'backup_url' - how to import the backup, and
        :returns: 'backup_service' describing the needed driver.
        :raises InvalidBackup:
        """
        LOG.info('Export record started, backup: %s.', backup.id)

        expected_status = fields.BackupStatus.AVAILABLE
        actual_status = backup.status
        if actual_status != expected_status:
            err = (_('Export backup aborted, expected backup status '
                     '%(expected_status)s but got %(actual_status)s.') %
                   {'expected_status': expected_status,
                    'actual_status': actual_status})
            raise exception.InvalidBackup(reason=err)

        backup_record = {'backup_service': backup.service}
        if not self._is_our_backup(backup):
            err = (_('Export record aborted, the backup service currently '
                     'configured [%(configured_service)s] is not the '
                     'backup service that was used to create this '
                     'backup [%(backup_service)s].') %
                   {'configured_service': self.driver_name,
                    'backup_service': backup.service})
            raise exception.InvalidBackup(reason=err)

        # Call driver to create backup description string
        try:
            backup_service = self.service(context)
            driver_info = backup_service.export_record(backup)
            backup_url = backup.encode_record(driver_info=driver_info)
            backup_record['backup_url'] = backup_url
        except Exception as err:
            msg = str(err)
            raise exception.InvalidBackup(reason=msg)

        LOG.info('Export record finished, backup %s exported.', backup.id)
        return backup_record

    def import_record(self,
                      context,
                      backup,
                      backup_service,
                      backup_url,
                      backup_hosts):
        """Import all volume backup metadata details to the backup db.

        :param context: running context
        :param backup: The new backup object for the import
        :param backup_service: The needed backup driver for import
        :param backup_url: An identifier string to locate the backup
        :param backup_hosts: Potential hosts to execute the import
        :raises InvalidBackup:
        :raises ServiceNotFound:
        """
        LOG.info('Import record started, backup_url: %s.', backup_url)

        # Can we import this backup?
        if not self._is_our_backup(backup_service):
            # No, are there additional potential backup hosts in the list?
            if len(backup_hosts) > 0:
                # try the next host on the list, maybe he can import
                first_host = backup_hosts.pop()
                self.backup_rpcapi.import_record(context,
                                                 first_host,
                                                 backup,
                                                 backup_service,
                                                 backup_url,
                                                 backup_hosts)
            else:
                # empty list - we are the last host on the list, fail
                err = _('Import record failed, cannot find backup '
                        'service to perform the import. Request service '
                        '%(service)s.') % {'service': backup_service}
                volume_utils.update_backup_error(backup, err)
                raise exception.ServiceNotFound(service_id=backup_service)
        else:
            # Yes...
            try:
                # Deserialize backup record information
                backup_options = backup.decode_record(backup_url)

                # Extract driver specific info and pass it to the driver
                driver_options = backup_options.pop('driver_info', {})
                backup_service = self.service(context)
                backup_service.import_record(backup, driver_options)
            except Exception as err:
                msg = str(err)
                volume_utils.update_backup_error(backup, msg)
                raise exception.InvalidBackup(reason=msg)

            required_import_options = {
                'display_name',
                'display_description',
                'container',
                'size',
                'service_metadata',
                'object_count',
                'id'
            }

            # Check for missing fields in imported data
            missing_opts = required_import_options - set(backup_options)
            if missing_opts:
                msg = (_('Driver successfully decoded imported backup data, '
                         'but there are missing fields (%s).') %
                       ', '.join(missing_opts))
                volume_utils.update_backup_error(backup, msg)
                raise exception.InvalidBackup(reason=msg)

            # Confirm the ID from the record in the DB is the right one
            backup_id = backup_options['id']
            if backup_id != backup.id:
                msg = (_('Trying to import backup metadata from id %(meta_id)s'
                         ' into backup %(id)s.') %
                       {'meta_id': backup_id, 'id': backup.id})
                volume_utils.update_backup_error(backup, msg)
                raise exception.InvalidBackup(reason=msg)

            # Overwrite some fields
            backup_options['service'] = self.driver_name
            backup_options['availability_zone'] = self.az
            backup_options['host'] = self.host

            # Remove some values which are not actual fields and some that
            # were set by the API node
            for key in ('name', 'user_id', 'project_id', 'deleted_at',
                        'deleted', 'fail_reason', 'status'):
                backup_options.pop(key, None)

            # Update the database
            backup.update(backup_options)
            backup.save()

            # Update the backup's status
            backup.update({"status": fields.BackupStatus.AVAILABLE})
            backup.save()

            LOG.info('Import record id %s metadata from driver '
                     'finished.', backup.id)

    def reset_status(self, context, backup, status):
        """Reset volume backup status.

        :param context: running context
        :param backup: The backup object for reset status operation
        :param status: The status to be set
        :raises InvalidBackup:
        :raises AttributeError:
        """
        LOG.info('Reset backup status started, backup_id: '
                 '%(backup_id)s, status: %(status)s.',
                 {'backup_id': backup.id,
                  'status': status})

        LOG.info('Backup service: %s.', backup.service)
        if not self._is_our_backup(backup):
            err = _('Reset backup status aborted, the backup service'
                    ' currently configured [%(configured_service)s] '
                    'is not the backup service that was used to create'
                    ' this backup [%(backup_service)s].') % \
                {'configured_service': self.driver_name,
                 'backup_service': backup.service}
            raise exception.InvalidBackup(reason=err)

        if backup.service is not None:
            backup.status = status
            backup.save()

            # Needs to clean temporary volumes and snapshots.
            try:
                self._cleanup_temp_volumes_snapshots_for_one_backup(
                    context, backup)
            except Exception:
                LOG.exception("Problem cleaning temp volumes and "
                              "snapshots for backup %(bkup)s.",
                              {'bkup': backup.id})

            volume_utils.notify_about_backup_usage(context, backup,
                                                   'reset_status.end')

    def check_support_to_force_delete(self, context):
        """Check if the backup driver supports force delete operation.

        :param context: running context
        """
        backup_service = self.service(context)
        return backup_service.support_force_delete

    def _attach_device(self, ctxt, backup_device,
                       properties, is_snapshot=False):
        """Attach backup device."""
        if not is_snapshot:
            return self._attach_volume(ctxt, backup_device, properties)
        else:
            return self._attach_snapshot(ctxt, backup_device, properties)

    def _attach_volume(self, context, volume, properties):
        """Attach a volume."""

        try:
            conn = self.volume_rpcapi.initialize_connection(context,
                                                            volume,
                                                            properties)
            return self._connect_device(conn)
        except Exception:
            with excutils.save_and_reraise_exception():
                try:
                    self.volume_rpcapi.terminate_connection(context, volume,
                                                            properties,
                                                            force=True)
                except Exception:
                    LOG.warning("Failed to terminate the connection "
                                "of volume %(volume_id)s, but it is "
                                "acceptable.",
                                {'volume_id': volume.id})

    def _attach_snapshot(self, ctxt, snapshot, properties):
        """Attach a snapshot."""

        try:
            conn = self.volume_rpcapi.initialize_connection_snapshot(
                ctxt, snapshot, properties)
            return self._connect_device(conn)
        except Exception:
            with excutils.save_and_reraise_exception():
                try:
                    self.volume_rpcapi.terminate_connection_snapshot(
                        ctxt, snapshot, properties, force=True)
                except Exception:
                    LOG.warning("Failed to terminate the connection "
                                "of snapshot %(snapshot_id)s, but it is "
                                "acceptable.",
                                {'snapshot_id': snapshot.id})

    def _connect_device(self, conn):
        """Establish connection to device."""
        use_multipath = CONF.use_multipath_for_image_xfer
        device_scan_attempts = CONF.num_volume_device_scan_tries
        protocol = conn['driver_volume_type']
        connector = volume_utils.brick_get_connector(
            protocol,
            use_multipath=use_multipath,
            device_scan_attempts=device_scan_attempts,
            conn=conn,
            expect_raw_disk=True)
        vol_handle = connector.connect_volume(conn['data'])

        return {'conn': conn, 'device': vol_handle, 'connector': connector}

    def _detach_device(self, ctxt, attach_info, device,
                       properties, is_snapshot=False, force=False,
                       ignore_errors=False):
        """Disconnect the volume or snapshot from the host. """
        connector = attach_info['connector']
        connector.disconnect_volume(attach_info['conn']['data'],
                                    attach_info['device'],
                                    force=force, ignore_errors=ignore_errors)
        rpcapi = self.volume_rpcapi
        if not is_snapshot:
            rpcapi.terminate_connection(ctxt, device, properties,
                                        force=force)
            rpcapi.remove_export(ctxt, device, sync=True)
        else:
            rpcapi.terminate_connection_snapshot(ctxt, device,
                                                 properties, force=force)
            rpcapi.remove_export_snapshot(ctxt, device, sync=True)

    def is_working(self):
        return self.is_initialized

    @periodic_task.periodic_task(
        spacing=CONF.backup_driver_stats_polling_interval)
    def publish_service_capabilities(self, context):
        """Collect driver status and then publish."""
        self._report_driver_status(context)
        self._publish_service_capabilities(context)

    def _report_driver_status(self, context):
        backup_stats = {
            'backend_state': self.is_working(),
            'driver_name': self.driver_name,
            'availability_zone': self.az
        }
        self.update_service_capabilities(backup_stats)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from cinder import context
from cinder import db
from cinder import exception
from cinder import objects
from cinder.tests.unit.api.v2 import fakes as v2_fakes
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import test
from cinder.tests.unit import utils as tests_utils
from cinder.volume import api as volume_api
from cinder.volume import configuration as conf


class AttachmentManagerTestCase(test.TestCase):
    """Attachment related test for volume/api.py."""

    def setUp(self):
        """Setup test class."""
        super(AttachmentManagerTestCase, self).setUp()
        self.configuration = mock.Mock(conf.Configuration)
        self.context = context.get_admin_context()
        self.context.user_id = fake.USER_ID
        self.project_id = fake.PROJECT3_ID
        self.context.project_id = self.project_id
        self.volume_api = volume_api.API()
        self.user_context = context.RequestContext(
            user_id=fake.USER_ID,
            project_id=fake.PROJECT3_ID)

    def test_attachment_create_no_connector(self):
        """Test attachment_create no connector."""
        volume_params = {'status': 'available'}

        vref = tests_utils.create_volume(self.context, **volume_params)
        aref = self.volume_api.attachment_create(self.context,
                                                 vref,
                                                 fake.UUID2)
        self.assertEqual(fake.UUID2, aref.instance_uuid)
        self.assertIsNone(aref.attach_time)
        self.assertEqual('reserved', aref.attach_status)
        self.assertEqual('null', aref.attach_mode)
        self.assertEqual(vref.id, aref.volume_id)
        self.assertEqual({}, aref.connection_info)

    @mock.patch('cinder.volume.rpcapi.VolumeAPI.attachment_update')
    def test_attachment_create_with_connector(self,
                                              mock_rpc_attachment_update):
        """Test attachment_create with connector."""
        volume_params = {'status': 'available'}
        connection_info = {'fake_key': 'fake_value',
                           'fake_key2': ['fake_value1', 'fake_value2']}
        mock_rpc_attachment_update.return_value = connection_info

        vref = tests_utils.create_volume(self.context, **volume_params)
        connector = {'fake': 'connector'}
        attachment = self.volume_api.attachment_create(self.context,
                                                       vref,
                                                       fake.UUID2,
                                                       connector)
        mock_rpc_attachment_update.assert_called_once_with(self.context,
                                                           mock.ANY,
                                                           connector,
                                                           mock.ANY)
        new_attachment = objects.VolumeAttachment.get_by_id(self.context,
                                                            attachment.id)
        self.assertEqual(connection_info, new_attachment.connection_info)

    @mock.patch('cinder.volume.rpcapi.VolumeAPI.attachment_delete')
    def test_attachment_delete_reserved(self,
                                        mock_rpc_attachment_delete):
        """Test attachment_delete with reserved."""
        volume_params = {'status': 'available'}

        vref = tests_utils.create_volume(self.context, **volume_params)
        aref = self.volume_api.attachment_create(self.context,
                                                 vref,
                                                 fake.UUID2)
        aobj = objects.VolumeAttachment.get_by_id(self.context,
                                                  aref.id)
        self.assertEqual('reserved', aref.attach_status)
        self.assertEqual(vref.id, aref.volume_id)
        self.volume_api.attachment_delete(self.context,
                                          aobj)

        # Since it's just reserved and never finalized, we should never make an
        # rpc call
        mock_rpc_attachment_delete.assert_not_called()

    @mock.patch('cinder.volume.rpcapi.VolumeAPI.attachment_delete')
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.attachment_update')
    def test_attachment_create_update_and_delete(
            self,
            mock_rpc_attachment_update,
            mock_rpc_attachment_delete):
        """Test attachment_delete."""
        volume_params = {'status': 'available'}
        connection_info = {'fake_key': 'fake_value',
                           'fake_key2': ['fake_value1', 'fake_value2']}
        mock_rpc_attachment_update.return_value = connection_info

        vref = tests_utils.create_volume(self.context, **volume_params)
        aref = self.volume_api.attachment_create(self.context,
                                                 vref,
                                                 fake.UUID2)
        aref = objects.VolumeAttachment.get_by_id(self.context,
                                                  aref.id)
        vref = objects.Volume.get_by_id(self.context,
                                        vref.id)

        connector = {'fake': 'connector',
                     'host': 'somehost'}
        self.volume_api.attachment_update(self.context,
                                          aref,
                                          connector)
        aref = objects.VolumeAttachment.get_by_id(self.context,
                                                  aref.id)
        self.assertEqual(connection_info, aref.connection_info)
        # We mock the actual call that updates the status
        # so force it here
        values = {'volume_id': vref.id,
                  'volume_host': vref.host,
                  'attach_status': 'attached',
                  'instance_uuid': fake.UUID2}
        aref = db.volume_attach(self.context, values)

        aref = objects.VolumeAttachment.get_by_id(self.context,
                                                  aref.id)
        self.assertEqual(vref.id, aref.volume_id)
        self.volume_api.attachment_delete(self.context,
                                          aref)

        mock_rpc_attachment_delete.assert_called_once_with(self.context,
                                                           aref.id,
                                                           mock.ANY)

    def test_additional_attachment_create_no_connector(self):
        """Test attachment_create no connector."""
        volume_params = {'status': 'available'}

        vref = tests_utils.create_volume(self.context, **volume_params)
        aref = self.volume_api.attachment_create(self.context,
                                                 vref,
                                                 fake.UUID2)
        self.assertEqual(fake.UUID2, aref.instance_uuid)
        self.assertIsNone(aref.attach_time)
        self.assertEqual('reserved', aref.attach_status)
        self.assertEqual('null', aref.attach_mode)
        self.assertEqual(vref.id, aref.volume_id)
        self.assertEqual({}, aref.connection_info)

        self.assertRaises(exception.InvalidVolume,
                          self.volume_api.attachment_create,
                          self.context,
                          vref,
                          fake.UUID1)
        self.volume_api.attachment_create(self.context,
                                          vref,
                                          fake.UUID2)
        vref = objects.Volume.get_by_id(self.context,
                                        vref.id)
        self.assertEqual(2, len(vref.volume_attachment))

    @mock.patch('cinder.volume.rpcapi.VolumeAPI.attachment_update')
    def test_attachment_create_reserve_delete(
            self,
            mock_rpc_attachment_update):
        volume_params = {'status': 'available'}
        connector = {
            "initiator": "iqn.1993-08.org.debian:01:cad181614cec",
            "ip": "192.168.1.20",
            "platform": "x86_64",
            "host": "tempest-1",
            "os_type": "linux2",
            "multipath": False}

        connection_info = {'fake_key': 'fake_value',
                           'fake_key2': ['fake_value1', 'fake_value2']}
        mock_rpc_attachment_update.return_value = connection_info

        vref = tests_utils.create_volume(self.context, **volume_params)
        aref = self.volume_api.attachment_create(self.context,
                                                 vref,
                                                 fake.UUID2,
                                                 connector=connector)
        vref = objects.Volume.get_by_id(self.context,
                                        vref.id)
        # Need to set the status here because our mock isn't doing it for us
        vref.status = 'in-use'
        vref.save()

        # Now a second attachment acting as a reserve
        self.volume_api.attachment_create(self.context,
                                          vref,
                                          fake.UUID2)

        # We should now be able to delete the original attachment that gave us
        # 'in-use' status, and in turn we should revert to the outstanding
        # attachments reserve
        self.volume_api.attachment_delete(self.context,
                                          aref)
        vref = objects.Volume.get_by_id(self.context,
                                        vref.id)
        self.assertEqual('reserved', vref.status)

    def test_reserve_reserve_delete(self):
        """Test that we keep reserved status across multiple reserves."""
        volume_params = {'status': 'available'}

        vref = tests_utils.create_volume(self.context, **volume_params)
        aref = self.volume_api.attachment_create(self.context,
                                                 vref,
                                                 fake.UUID2)
        vref = objects.Volume.get_by_id(self.context,
                                        vref.id)
        self.assertEqual('reserved', vref.status)

        self.volume_api.attachment_create(self.context,
                                          vref,
                                          fake.UUID2)
        vref = objects.Volume.get_by_id(self.context,
                                        vref.id)
        self.assertEqual('reserved', vref.status)
        self.volume_api.attachment_delete(self.context,
                                          aref)
        vref = objects.Volume.get_by_id(self.context,
                                        vref.id)
        self.assertEqual('reserved', vref.status)
        self.assertEqual(1, len(vref.volume_attachment))

    def test_attachment_create_readonly_volume(self):
        """Test attachment_create on a readonly volume."""
        volume_params = {'status': 'available'}

        vref = tests_utils.create_volume(self.context, **volume_params)
        self.volume_api.update_readonly_flag(self.context, vref, True)
        aref = self.volume_api.attachment_create(self.context,
                                                 vref,
                                                 fake.UUID2)
        self.assertEqual(fake.UUID2, aref.instance_uuid)
        self.assertIsNone(aref.attach_time)
        self.assertEqual('reserved', aref.attach_status)
        self.assertEqual('ro', aref.attach_mode)
        self.assertEqual(vref.id, aref.volume_id)
        self.assertEqual({}, aref.connection_info)

    def test_attachment_create_volume_in_error_state(self):
        """Test attachment_create volume in error state."""
        volume_params = {'status': 'available'}

        vref = tests_utils.create_volume(self.context, **volume_params)
        vref.status = "error"
        self.assertRaises(exception.InvalidVolume,
                          self.volume_api.attachment_create,
                          self.context,
                          vref,
                          fake.UUID2)

    def test_attachment_update_volume_in_error_state(self):
        """Test attachment_update volumem in error state."""
        volume_params = {'status': 'available'}

        vref = tests_utils.create_volume(self.context, **volume_params)
        aref = self.volume_api.attachment_create(self.context,
                                                 vref,
                                                 fake.UUID2)
        self.assertEqual(fake.UUID2, aref.instance_uuid)
        self.assertIsNone(aref.attach_time)
        self.assertEqual('reserved', aref.attach_status)
        self.assertEqual(vref.id, aref.volume_id)
        self.assertEqual({}, aref.connection_info)
        vref.status = 'error'
        vref.save()
        connector = {'fake': 'connector',
                     'host': 'somehost'}
        self.assertRaises(exception.InvalidVolume,
                          self.volume_api.attachment_update,
                          self.context,
                          aref,
                          connector)

    @mock.patch('cinder.db.sqlalchemy.api.volume_attachment_update',
                return_value={})
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.attachment_update',
                return_value={})
    @mock.patch.object(db.sqlalchemy.api, '_volume_type_get',
                       v2_fakes.fake_volume_type_get)
    def test_attachment_update_duplicate(self, mock_va_update, mock_db_upd):
        volume_params = {'status': 'available'}

        vref = tests_utils.create_volume(self.context,
                                         deleted=0,
                                         **volume_params)

        tests_utils.attach_volume(self.context,
                                  vref.id,
                                  fake.UUID1,
                                  'somehost',
                                  'somemountpoint')

        # Update volume with another attachment
        tests_utils.attach_volume(self.context,
                                  vref.id,
                                  fake.UUID2,
                                  'somehost2',
                                  'somemountpoint2')
        vref.refresh()

        # This attachment will collide with the first
        connector = {'host': 'somehost'}
        vref.volume_attachment[0]['connector'] = {'host': 'somehost'}
        vref.volume_attachment[0]['connection_info'] = {'c': 'd'}
        with mock.patch('cinder.objects.Volume.get_by_id', return_value=vref):
            with mock.patch.object(self.volume_api.volume_rpcapi,
                                   'attachment_update') as m_au:
                self.assertRaises(exception.InvalidVolume,
                                  self.volume_api.attachment_update,
                                  self.context,
                                  vref.volume_attachment[1],
                                  connector)
                m_au.assert_not_called()
        mock_va_update.assert_not_called()
        mock_db_upd.assert_not_called()

    def test_attachment_create_creating_volume(self):
        """Test attachment_create on a creating volume."""
        volume_params = {'status': 'creating'}

        vref = tests_utils.create_volume(self.context, **volume_params)
        self.assertRaises(exception.InvalidVolume,
                          self.volume_api.attachment_create,
                          self.context,
                          vref,
                          fake.UUID1)



# Copyright (c) 2018 Hedvig, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Volume driver for Hedvig Block Storage.

"""

import socket

from oslo_log import log as logging
from oslo_utils import strutils
from oslo_utils import units

from cinder.common import constants
from cinder import exception
from cinder.i18n import _
from cinder import interface
from cinder.volume import driver
from cinder.volume.drivers.hedvig import config
from cinder.volume.drivers.hedvig import rest_client
from cinder.volume.drivers.san import san
from cinder.volume import volume_types

LOG = logging.getLogger(__name__)


@interface.volumedriver
class HedvigISCSIDriver(driver.ISCSIDriver, san.SanDriver):
    """OpenStack Cinder driver to enable Hedvig storage.

    .. code-block:: none

     Version history:

        1.0 - Initial driver

    """
    DEFAULT_VOL_BLOCK_SIZE = 4 * units.Ki
    DEFAULT_CREATEDBY = "OpenStack"
    DEFAULT_EXPORT_BLK_SIZE = 4096
    DEFAULT_CAPACITY = units.Gi
    DEFAULT_ISCSI_PORT = 3260
    DEFAULT_TARGET_NAME = "iqn.2012-05.com.hedvig:storage."
    VERSION = "1.0.0"
    CI_WIKI_NAME = "Hedvig_CI"

    def __init__(self, *args, **kwargs):
        super(HedvigISCSIDriver, self).__init__(*args, **kwargs)
        self.group_stats = {}
        self.hrs = None

    @staticmethod
    def get_driver_options():
        return []

    def check_for_setup_error(self):
        self.hrs.connect()
        LOG.info("Initialization complete")

    def do_setup(self, context):
        # Ensure that the data required by hedvig are provided
        required_config = ['san_login', 'san_password', 'san_ip',
                           'san_clustername']
        for attr in required_config:
            if not getattr(self.configuration, attr, None):
                msg = _('Hedvig param %s is not set.') % attr
                LOG.error(msg)
                raise exception.VolumeDriverException(msg)
        self.san_ip = self.configuration.san_ip
        self.san_login = self.configuration.san_login
        self.san_password = self.configuration.san_password
        self.san_clustername = self.configuration.san_clustername
        LOG.info('Initializing hedvig cinder driver with '
                 'server: %s', self.san_ip)
        self.hrs = rest_client.RestClient(self.san_ip,
                                          self.san_login,
                                          self.san_password,
                                          self.san_clustername)

    def get_volume_stats(self, refresh=False):
        # we need to get stats for server.
        if refresh is True:
            total_capacity, free_capacity = self.update_volume_stats()
            stats = dict()
            stats["volume_backend_name"] = "hedvig"
            stats["vendor_name"] = "Hedvig Inc"
            stats["driver_version"] = self.VERSION
            stats["storage_protocol"] = constants.ISCSI
            stats["total_capacity_gb"] = total_capacity
            stats["free_capacity_gb"] = free_capacity
            stats["QoS_support"] = True
            self.group_stats = stats
        return self.group_stats

    def create_volume(self, volume):
        """Driver entry point for creating a new volume."""
        try:
            qos_specs = None
            name, description, size = self.get_hedvig_volume_details(volume)
            vol_type_id = volume.volume_type_id
            if vol_type_id is not None:
                qos = volume_types.get_volume_type_qos_specs(vol_type_id)
                qos_specs = qos['qos_specs']
            self.hedvig_create_virtualdisk(name, description, size, qos_specs)
        except exception.VolumeDriverException:
            msg = _('Failed to create volume %s. Rest API failed'
                    ) % volume.name
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)
        except Exception:
            msg = _('Failed to create volume: %s') % volume.name
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)

    def delete_volume(self, volume):
        """Driver entry point for deleting volume."""
        LOG.debug("Deleting volume: %s", volume.name)
        name = volume.name
        try:
            self.hedvig_delete_virtualdisk(name)
        except exception.VolumeDriverException:
            msg = _('Failed to delete volume %s. Rest API failed'
                    ) % volume.name
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)
        except Exception:
            msg = _('Failed to delete volume: %s') % volume.name
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)

    def create_cloned_volume(self, volume, src_vref):
        """Create a clone of the volume."""
        try:
            LOG.debug('Create cloned volume called '
                      'volume_id = %(volume)s and src_vol_id = %(src_vol_id)s',
                      {'volume': volume.id, 'src_vol_id': src_vref.id})
            name, desc, size = self.get_hedvig_volume_details(volume)
            self.hrs.clone_vdisk(srcVolName=src_vref.name, dstVolName=name,
                                 size=size)
        except exception.VolumeDriverException:
            msg = _('Failed to create cloned volume. Rest API failed')
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)
        except Exception:
            msg = _('Failed to create cloned volume')
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)

    def initialize_connection(self, volume, connector):
        """Driver entry point to attach a volume to an instance.

        Assign any created volume to a compute node/controllerVM so
        that it can be attached to a instance.
        This driver returns a driver_volume_type of 'iscsi'.
        The format of the driver data is defined as follows -- similar
        to _get_iscsi_properties.
        """
        LOG.debug('Initializing connection. volume: %s, '
                  'connector: %s', volume, connector)
        try:
            computeHost = self.get_compute_host(connector)
            volName = volume.name
            tgtHost = self.hedvig_lookup_tgt(computeHost)
            if tgtHost is None:
                LOG.warning("No target registered for compute host %s",
                            computeHost)
                tgtHost = self.hedvig_lookup_tgt()
            lunnum = self.hedvig_get_lun(tgtHost, volName)
            if lunnum == -1:
                LOG.error('Failed to get lun for volume: %s, '
                          'hedvig controller: %s', volume, tgtHost)
                raise exception.VolumeDriverException()

            # Add access to the mgmt interface addr and iqn of compute host
            LOG.debug("Calling add access %(host)s : %(vol)s : %(iqn)s ",
                      {'host': tgtHost, 'vol': volName,
                       'iqn': connector['initiator']})
            self.hedvig_add_access(tgtHost, volName, connector['initiator'])

            # Add access to both storage and mgmt interface addrs for
            # iscsi discovery to succeed
            LOG.debug("Calling hedvig_get_iqn %s", socket.getfqdn())
            controller_host_iqn = self.hedvig_get_iqn(socket.getfqdn())

            LOG.debug("Calling add access with %s : %s : %s ", tgtHost,
                      volName, controller_host_iqn)
            self.hedvig_add_access(tgtHost, volName, controller_host_iqn)
            targetName = ("%s%s-%s" % (self.DEFAULT_TARGET_NAME, tgtHost,
                                       lunnum))
            portal = ("%s:%s" % (socket.gethostbyname(tgtHost),
                                 self.DEFAULT_ISCSI_PORT))
            iscsi_properties = ({'target_discovered': True,
                                 'target_iqn': targetName,
                                 'target_portal': portal,
                                 'target_lun': lunnum})
            LOG.debug("iscsi_properties: %s", iscsi_properties)
            return {'driver_volume_type': 'iscsi', 'data': iscsi_properties}
        except exception.VolumeDriverException:
            msg = _('Volume assignment to connect failed. volume: %s '
                    'Rest API failed') % volume
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)
        except Exception:
            msg = _('Volume assignment to connect failed. volume: %s') % volume
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)

    def terminate_connection(self, volume, connector, **kwargs):
        """Driver entry point to detach volume from instance."""
        LOG.debug("Terminating connection. volume: %s, connector: %s",
                  volume, connector)
        try:
            volName = volume.name
            if connector is None:
                LOG.debug("Removing ALL host connections for volume %s",
                          volume)
                targetList = self.hrs.list_targets(computeHost=None)
                for target in targetList:
                    self.hedvig_delete_lun(target, volName)
                return
            computeHost = self.get_compute_host(connector)
            tgtHost = self.hedvig_lookup_tgt(computeHost)
            if tgtHost is None:
                LOG.debug("No target registered for compute host %s",
                          computeHost)
                tgtHost = self.hedvig_lookup_tgt()
            if tgtHost is None:
                msg = _('Failed to get hedvig controller')
                LOG.error(msg)
                raise exception.VolumeDriverException(msg)
            self.hedvig_delete_lun(tgtHost, volName)
        except exception.VolumeDriverException:
            msg = _('Failed to terminate connection. volume: %s '
                    'Rest API failed') % volume
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)
        except Exception:
            msg = _('Failed to terminate connection. volume: %s') % volume
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)

    def create_snapshot(self, snapshot):
        """Driver entry point for creating a snapshot."""
        try:
            volName = snapshot.volume_name
            snapshotName = snapshot.name
            project = snapshot.project_id
            snapshotId = snapshot.id
            LOG.info("Creating snapshot. volName: %s, snapshotName: %s, "
                     "project: %s, snapshotId: %s", volName,
                     snapshotName, project, snapshotId)
            self.hedvig_create_snapshot(volName, snapshotId)
        except exception.VolumeDriverException:
            msg = (_('Failed to create snapshot. snapshotName: %s '
                     'Rest API failed') % snapshotName)
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)
        except Exception:
            msg = (_('Failed to create snapshot. snapshotName: %s')
                   % snapshotName)
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)

    def delete_snapshot(self, snapshot):
        """Driver entry point for deleting a snapshot."""
        try:
            volName = snapshot.volume_name
            snapshotName = snapshot.display_name
            project = snapshot.project_id
            snapshotId = snapshot.id
            LOG.info("Deleting snapshot. volName: %s, snapshotName: %s, "
                     "project: %s", volName, snapshotName, project)
            self.hrs.delete_snapshot(snapshotName, volName, snapshotId)
        except exception.VolumeDriverException:
            msg = _('Failed to delete snapshot: %s, '
                    'Rest API failed') % snapshotName
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)
        except Exception:
            msg = _('Failed to delete snapshot: %s') % snapshotName
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)

    def create_volume_from_snapshot(self, volume, snapshot):
        """Driver entry point for creating a new volume from a snapshot.

        This is the same as cloning.
        """
        name, description, size = self.get_hedvig_volume_details(volume)
        snapshotName = snapshot.display_name
        snapshotId = snapshot.id
        srcVolName = snapshot.volume_name
        try:
            LOG.info('Creating volume from snapshot. Name: %(volname)s,'
                     ' SrcVolName: %(src)s, Snap_id: %(sid)s',
                     {'volname': name, 'src': srcVolName, 'sid': snapshotId})
            self.hedvig_clone_snapshot(name, snapshotId, srcVolName, size)
        except exception.VolumeDriverException:
            msg = _('Failed to create volume from snapshot %s'
                    ' Rest API failed') % snapshotName
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)
        except Exception:
            msg = _('Failed to create volume from snapshot %s') % snapshotName
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)

    def extend_volume(self, volume, newSize):
        """Resizes virtual disk.

        newSize should be greater than current size.
        """
        try:
            name, description, size = self.get_hedvig_volume_details(volume)
            LOG.info('Resizing virtual disk. name: %s, '
                     'newSize: %s', name, newSize)
            if (size / units.Gi) >= newSize:
                err = _("Shrinking of volumes are not allowed")
                LOG.error(err)
                raise exception.VolumeDriverException(err)
            self.hrs.resize_vdisk(
                name,
                newSize)
        except exception.VolumeDriverException:
            msg = _('Failed to extend volume. Rest API failed')
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)
        except Exception:
            msg = _('Failed to extend volume')
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)

    def check_for_export(self, context, volume_id):
        """Not relevant to Hedvig"""
        pass

    def get_export(self, volume):
        """Get the iSCSI export details for a volume."""
        pass

    def ensure_export(self, context, volume):
        """Driver entry point to get the export info for an existing volume.

        Irrelevant for Hedvig. Export is created during attachment to instance.
        """
        pass

    def create_export(self, context, volume, properties):
        """Driver entry point to get the export info for a new volume.

        Irrelevant for Hedvig. Export is created during attachment to instance.
        """
        pass

    def remove_export(self, context, volume):
        """Driver entry point to remove an export for a volume.

        Irrelevant for Hedvig. Export should be deleted on detachment.
        """
        pass

    def detach_volume(self, context, volume, attachment):
        pass

    def hedvig_create_snapshot(self, vDiskName, snapshotId=None):
        """Hedvig call to create snapshot of vdisk."""
        LOG.debug("Creating snapshot..%s , %s.", vDiskName, snapshotId)
        try:
            snapshotName = self.hrs.create_snapshot(vDiskName, snapshotId)
            LOG.debug("Received snapshotName %s from rest call",
                      snapshotName)
            return snapshotName
        except exception.VolumeDriverException:
            msg = _('Failed to create snapshot for vdisk %s '
                    'Rest API failed') % vDiskName
            LOG.exception(msg)
            raise exception.VolumeDriverException()
        except Exception:
            msg = _('Failed to create snapshot for vdisk %s') % vDiskName
            LOG.exception(msg)
            raise exception.VolumeDriverException()

    def update_volume_stats(self):
        LOG.debug('Update volume stats called')
        try:
            total_capacity, free_capacity = self.hrs.update_volume_stats()
        except exception.VolumeDriverException:
            msg = _('Unable to fetch volume stats. Rest API failed')
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)
        except Exception:
            msg = _('Unable to fetch volume stats')
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)
        return (total_capacity, free_capacity)

    def get_hedvig_volume_details(self, volume):
        volName = volume.name
        project = volume.project_id
        displayName = volume.display_name
        displayDescription = volume.display_description
        description = ("%s\n%s\n%s" % (project, displayName,
                                       displayDescription))
        size = volume.size * units.Gi
        return volName, description, size

    def get_compute_host(self, connector):
        connectorHost = socket.getfqdn(connector['host'])
        localHost = socket.gethostname()
        computeHost = localHost
        if connectorHost != localHost:
            computeHost = connectorHost
        return computeHost

    def hedvig_lookup_tgt(self, host=None):
        """Get the tgt instance associated with the compute host"""
        LOG.debug("Looking up hedvig controller for compute host: %s",
                  host)
        try:
            targetList = self.hrs.list_targets(computeHost=host)
            tgt = None
            if len(targetList) > 0:
                tgt = targetList[0]

            LOG.debug("Found hedvig controller: %s, for host: %s", tgt, host)
            return tgt
        except exception.VolumeDriverException:
            msg = _('Failed to get hedvig controller for compute %s '
                    'Rest API failed') % host
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)
        except Exception:
            msg = _('Failed to get hedvig controller for compute %s ') % host
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)

    def hedvig_delete_lun(self, tgtHost, vDiskName):
        try:
            LOG.debug("Deleting lun. hedvig controller: %s, vDiskName: %s,",
                      tgtHost, vDiskName)
            self.hrs.unmap_lun(tgtHost, vDiskName)
        except Exception:
            msg = _('Failed to delete lun')
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)

    def hedvig_get_lun(self, tgtHost, vDiskName):
        """Looks up lun based on tgthost and vDiskName.

        If lun does not exist then call add_lun and return the lun number.
        If lun exists, just return the lun number.
        """
        LOG.debug("Getting lun. hedvig controller: %s, vDiskName: %s",
                  tgtHost, vDiskName)
        try:
            lunNo = self.hrs.get_lun(tgtHost, vDiskName)
            if lunNo > -1:
                return lunNo

            # If the lun is not found, add lun for the vdisk
            LOG.debug("Calling add lun on target : %s vdisk %s", tgtHost,
                      vDiskName)
            self.hrs.add_lun(tgtHost, vDiskName, False)
            lunNo = self.hrs.get_lun(tgtHost, vDiskName)
            return lunNo

        except Exception:
            msg = _('Failed to get lun for vdisk: %s') % vDiskName
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)

    def hedvig_get_iqn(self, hostname):
        """Looks up the iqn for the given host."""
        try:
            iqn = self.hrs.get_iqn(hostname)
            LOG.debug("Got IQN: %s, for hostname: %s", iqn, hostname)
            return iqn
        except Exception:
            msg = _('Failed to get iqn for hostname: %s') % hostname
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)

    def hedvig_add_access(self, tgtHost, volName, initiator):
        """Adds access to LUN for initiator's ip/iqn."""
        try:
            LOG.info("Adding access. hedvig controller: %s, vol name  %s, "
                     "initiator: %s", tgtHost, volName, initiator)
            self.hrs.add_access(tgtHost, volName, "iqn", initiator)
        except Exception:
            msg = _('Failed to add access. hedvig controller: %s') % tgtHost
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)

    def hedvig_create_virtualdisk(self, name, description, size, qos_specs):
        try:
            LOG.info('Creating virtual disk. name: %s, description: %s,'
                     'size: %s', name, description, size)
            vDiskInfo = {
                'name': name,
                'blockSize': HedvigISCSIDriver.DEFAULT_VOL_BLOCK_SIZE,
                'size': size,
                'createdBy':
                    HedvigISCSIDriver.DEFAULT_CREATEDBY,
                'description': description,
                'residence': config.Config.DiskResidence[1],
                'replicationFactor': 3,
                'replicationPolicy': 'Agnostic',
                'clusteredFileSystem': False,
                'exportedBlockSize': HedvigISCSIDriver.DEFAULT_EXPORT_BLK_SIZE,
                'cacheEnabled': config.Config.defaultCinderCacheEnable,
                'diskType': 'BLOCK',
                'immutable': False,
                'deduplication': config.Config.defaultCinderDedupEnable,
                'compressed': config.Config.defaultCinderCompressEnable,
                'cloudEnabled': False,
                'cloudProvider': 0,
                'isClone': False,
                'consistency': 'STRONG',
                'scsi3pr': False
            }
            if qos_specs:
                kvs = qos_specs['specs']
                for key, value in kvs.items():
                    if "dedup_enable" == key:
                        val = self.parse_and_get_boolean_entry(
                            value)
                        if val:
                            vDiskInfo['deduplication'] = val
                    elif "compressed_enable" == key:
                        val = self.parse_and_get_boolean_entry(
                            value)
                        if val:
                            vDiskInfo['compressed'] = True
                    elif "cache_enable" == key:
                        val = self.parse_and_get_boolean_entry(
                            value)
                        if val:
                            vDiskInfo['cacheEnabled'] = val
                    elif "encryption" == key:
                        val = self.parse_and_get_boolean_entry(
                            value)
                        if val:
                            vDiskInfo['encryption'] = val
                    elif "replication_factor" == key:
                        val = int(value)
                        if val > 0:
                            vDiskInfo['replicationFactor'] = val
                    elif "replication_policy" == key:
                        val = value.strip(" \n\t").lower()
                        if val:
                            vDiskInfo['replicationPolicy'] = val
                    elif "disk_residence" == key:
                        val = value.strip(" \n\t").lower()
                        if val:
                            vDiskInfo['residence'] = val
                    elif "replication_policy_info" == key:
                        val = value.split(',')
                        if len(val) != 0:
                            dcList = []
                            for dataCenter in val:
                                dcList.append(dataCenter)
                            vDiskInfo['dataCenters'] = dcList

            if vDiskInfo['deduplication'] and (
                    vDiskInfo['compressed'] is False):
                LOG.error('Cannot create dedup enabled disk without'
                          ' compression enabled')
                raise exception.VolumeDriverException()
            self.hrs.create_vdisk(vDiskInfo)
        except Exception:
            msg = _('Failed to create volume')
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)

    def hedvig_delete_virtualdisk(self, name):
        LOG.info('Deleting virtual disk. name - %s', name)
        try:
            self.hrs.delete_vdisk(name)
        except Exception:
            msg = _('Failed to delete Vdisk')
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)

    def hedvig_clone_snapshot(self, dstVolName,
                              openstackSID, srcVolName, size):
        LOG.info("Cloning a snapshot.dstVolName: %s,openstackSID:%s,"
                 "srcVolName: %s", dstVolName, openstackSID, srcVolName)
        try:
            self.hrs.clone_hedvig_snapshot(
                dstVolName=dstVolName,
                snapshotID=openstackSID,
                srcVolName=srcVolName,
                size=size)
        except Exception:
            msg = _('Failed to clone snapshot')
            LOG.exception(msg)
            raise exception.VolumeDriverException(msg)

    def parse_and_get_boolean_entry(self, entry):
        entry = entry.strip(" \t\n")
        return strutils.bool_from_string(entry)



# Copyright 2021 Red Hat, Inc.
# All Rights Reserved.
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt

from cinder.api.contrib import used_limits
from cinder.api import microversions as mv
from cinder.api.v3 import limits
from cinder.policies import limits as policy
from cinder.tests.unit.api import fakes as fake_api
from cinder.tests.unit.policies import base


@ddt.ddt
class LimitsPolicyTest(base.BasePolicyTest):
    authorized_readers = [
        'legacy_admin',
        'legacy_owner',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
        'project_foo',
        # The other_* users are allowed because we don't have any check
        # mechanism in the code to validate the project_id, which is
        # validated at the WSGI layer.
        'other_project_member',
        'other_project_reader',
    ]

    unauthorized_readers = [
        'system_member',
        'system_reader',
        'system_foo',
    ]

    unauthorized_exceptions = []

    # Basic policy test is without enforcing scope (which cinder doesn't
    # yet support) and deprecated rules enabled.
    def setUp(self, enforce_scope=False, enforce_new_defaults=False,
              *args, **kwargs):
        super().setUp(enforce_scope, enforce_new_defaults, *args, **kwargs)
        self.limits_controller = limits.LimitsController()
        self.used_limits_controller = used_limits.UsedLimitsController()
        self.api_path = '/v3/%s/limits' % (self.project_id)
        self.api_version = mv.BASE_VERSION

    @ddt.data(*base.all_users)
    def test_extend_limit_attribute_policy(self, user_id):
        rule_name = policy.EXTEND_LIMIT_ATTRIBUTE_POLICY
        url = self.api_path

        # Create a resp_obj (necessary for the UsedLimitsController) by
        # requesting the limits via the LimitsController, which actually
        # generates the response.
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.environ['cinder.context'] = self.project_admin_context
        limits = self.limits_controller.index(req)['limits']
        resp_obj = mock.MagicMock(obj={'limits': limits})

        # This proves the LimitsController's response doesn't include any
        # "used" entries (e.g. totalVolumesUsed).
        self.assertNotIn('totalVolumesUsed', limits['absolute'].keys())

        # Now hit the UsedLimitsController and see if it adds "used"
        # limits to the resp_obj.
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        self.common_policy_check(user_id, self.authorized_readers,
                                 self.unauthorized_readers,
                                 self.unauthorized_exceptions,
                                 rule_name,
                                 self.used_limits_controller.index,
                                 req, resp_obj=resp_obj, fatal=False)

        if user_id in self.authorized_readers:
            self.assertIn('totalVolumesUsed', limits['absolute'].keys())
        else:
            self.assertNotIn('totalVolumesUsed', limits['absolute'].keys())


class LimitsPolicySecureRbacTest(LimitsPolicyTest):
    authorized_readers = [
        'legacy_admin',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
        'other_project_member',
        'other_project_reader',
    ]

    unauthorized_readers = [
        'legacy_owner',
        'system_member',
        'system_reader',
        'system_foo',
        'project_foo',
    ]

    def setUp(self, *args, **kwargs):
        # Test secure RBAC by disabling deprecated policy rules (scope
        # is still not enabled).
        super().setUp(enforce_scope=False, enforce_new_defaults=True,
                      *args, **kwargs)



# Copyright (c) 2021 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock
import uuid

import ddt
import six

from cinder.tests.unit import test
from cinder.tests.unit.volume.drivers.dell_emc.powerstore import MockResponse
from cinder.volume.drivers.dell_emc.powerstore import client


CLIENT_OPTIONS = {
    "rest_ip": "127.0.0.1",
    "rest_username": "fake_user",
    "rest_password": "fake_password",
    "verify_certificate": False,
    "certificate_path": None
}

ISCSI_IP_POOL_RESP = [
    {
        "address": "1.2.3.4",
        "ip_port": {
            "target_iqn":
                "iqn.2022-07.com.dell:dellemc-powerstore-fake-iqn-1"
        },
    },
    {
        "address": "5.6.7.8",
        "ip_port": {
            "target_iqn":
                "iqn.2022-07.com.dell:dellemc-powerstore-fake-iqn-1"
        },
    },
]

NVME_IP_POOL_RESP = [
    {
        "address": "11.22.33.44"
    },
    {
        "address": "55.66.77.88"
    }
]


@ddt.ddt
class TestClient(test.TestCase):

    def setUp(self):
        super(TestClient, self).setUp()

        self.client = client.PowerStoreClient(**CLIENT_OPTIONS)
        self.fake_volume = six.text_type(uuid.uuid4())

    @ddt.data(("iSCSI", ISCSI_IP_POOL_RESP),
              ("NVMe", NVME_IP_POOL_RESP))
    @ddt.unpack
    @mock.patch("requests.request")
    def test_get_ip_pool_address(self, protocol, ip_pool, mock_request):
        mock_request.return_value = MockResponse(ip_pool, rc=200)
        response = self.client.get_ip_pool_address(protocol)
        mock_request.assert_called_once()
        self.assertEqual(response, ip_pool)

    @mock.patch("requests.request")
    def test_get_volume_nguid(self, mock_request):
        mock_request.return_value = MockResponse(
            content={
                "nguid": "nguid.76e02b0999y439958ttf546800ea7fe8"
            },
            rc=200
        )
        self.assertEqual(self.client.get_volume_nguid(self.fake_volume),
                         "76e02b0999y439958ttf546800ea7fe8")

    @mock.patch("requests.request")
    def test_get_array_version(self, mock_request):
        mock_request.return_value = MockResponse(
            content=[
                {
                    "release_version": "3.0.0.0",
                }
            ],
            rc=200
        )
        self.assertEqual(self.client.get_array_version(),
                         "3.0.0.0")



#    Copyright 2014 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import textwrap
from unittest import mock

import ddt
import pycodestyle

from cinder.tests.hacking import checks
from cinder.tests.unit import test


@ddt.ddt
class HackingTestCase(test.TestCase):
    """This class tests cinder's hacking checks.

    This class ensures that Cinder's hacking checks are working by passing
    strings to the check methods like the pep8/flake8 parser would. The parser
    loops over each line in the file and then passes the parameters to the
    check method. The parameter names in the check method dictate what type of
    object is passed to the check method. The parameter types are::

        logical_line: A processed line with the following modifications:
            - Multi-line statements converted to a single line.
            - Stripped left and right.
            - Contents of strings replaced with "xxx" of same length.
            - Comments removed.
        physical_line: Raw line of text from the input file.
        lines: a list of the raw lines from the input file
        tokens: the tokens that contribute to this logical line
        line_number: line number in the input file
        total_lines: number of lines in the input file
        blank_lines: blank lines before this one
        indent_char: indentation character in this file (" " or "\t")
        indent_level: indentation (with tabs expanded to multiples of 8)
        previous_indent_level: indentation on previous line
        previous_logical: previous logical line
        filename: Path of the file being run through pep8

    When running a test on a check method the return will be False/None if
    there is no violation in the sample input. If there is an error a tuple is
    returned with a position in the line, and a message. So to check the result
    just assertTrue if the check is expected to fail and assertFalse if it
    should pass.
    """

    def test_no_translate_logs(self):
        self.assertEqual(1, len(list(checks.no_translate_logs(
            "LOG.debug(_('foo'))", "cinder/scheduler/foo.py"))))
        self.assertEqual(1, len(list(checks.no_translate_logs(
            "LOG.error(_('foo'))", "cinder/scheduler/foo.py"))))
        self.assertEqual(1, len(list(checks.no_translate_logs(
            "LOG.info(_('foo'))", "cinder/scheduler/foo.py"))))
        self.assertEqual(1, len(list(checks.no_translate_logs(
            "LOG.warning(_('foo'))", "cinder/scheduler/foo.py"))))
        self.assertEqual(1, len(list(checks.no_translate_logs(
            "LOG.exception(_('foo'))", "cinder/scheduler/foo.py"))))
        self.assertEqual(1, len(list(checks.no_translate_logs(
            "LOG.critical(_('foo'))", "cinder/scheduler/foo.py"))))

    def test_check_explicit_underscore_import(self):
        self.assertEqual(1, len(list(checks.check_explicit_underscore_import(
            "LOG.info(_('My info message'))",
            "cinder.tests.unit/other_files.py"))))
        self.assertEqual(1, len(list(checks.check_explicit_underscore_import(
            "msg = _('My message')",
            "cinder.tests.unit/other_files.py"))))
        self.assertEqual(0, len(list(checks.check_explicit_underscore_import(
            "from cinder.i18n import _",
            "cinder.tests.unit/other_files.py"))))
        self.assertEqual(0, len(list(checks.check_explicit_underscore_import(
            "LOG.info(_('My info message'))",
            "cinder.tests.unit/other_files.py"))))
        self.assertEqual(0, len(list(checks.check_explicit_underscore_import(
            "msg = _('My message')",
            "cinder.tests.unit/other_files.py"))))
        self.assertEqual(0, len(list(checks.check_explicit_underscore_import(
            "from cinder.i18n import _",
            "cinder.tests.unit/other_files2.py"))))
        self.assertEqual(0, len(list(checks.check_explicit_underscore_import(
            "msg = _('My message')",
            "cinder.tests.unit/other_files2.py"))))
        self.assertEqual(0, len(list(checks.check_explicit_underscore_import(
            "_ = translations.ugettext",
            "cinder.tests.unit/other_files3.py"))))
        self.assertEqual(0, len(list(checks.check_explicit_underscore_import(
            "msg = _('My message')",
            "cinder.tests.unit/other_files3.py"))))
        # Complete code coverage by falling through all checks
        self.assertEqual(0, len(list(checks.check_explicit_underscore_import(
            "LOG.info('My info message')",
            "cinder.tests.unit/other_files4.py"))))
        self.assertEqual(1, len(list(checks.check_explicit_underscore_import(
            "msg = _('My message')",
            "cinder.tests.unit/other_files5.py"))))

    # We are patching pycodestyle/pep8 so that only the check under test is
    # actually installed.
    # TODO(eharney): don't patch private members of external libraries
    @mock.patch('pycodestyle._checks',
                {'physical_line': {}, 'logical_line': {}, 'tree': {}})
    def _run_check(self, code, checker, filename=None):
        pycodestyle.register_check(checker)

        lines = textwrap.dedent(code).strip().splitlines(True)

        checker = pycodestyle.Checker(filename=filename, lines=lines)
        checker.check_all()
        checker.report._deferred_print.sort()
        return checker.report._deferred_print

    def _assert_has_errors(self, code, checker, expected_errors=None,
                           filename=None):
        actual_errors = [e[:3] for e in
                         self._run_check(code, checker, filename)]
        self.assertEqual(expected_errors or [], actual_errors)

    def _assert_has_no_errors(self, code, checker, filename=None):
        self._assert_has_errors(code, checker, filename=filename)

    def test_logging_format_args(self):
        checker = checks.CheckLoggingFormatArgs
        code = """
               import logging
               LOG = logging.getLogger()
               LOG.info("Message without a second argument.")
               LOG.critical("Message with %s arguments.", 'two')
               LOG.debug("Volume %s caught fire and is at %d degrees C and"
                         " climbing.", 'volume1', 500)
               """
        self._assert_has_no_errors(code, checker)

        code = """
               import logging
               LOG = logging.getLogger()
               LOG.{0}("Volume %s caught fire and is at %d degrees C and "
                      "climbing.", ('volume1', 500))
               """
        # We don't assert on specific column numbers since there is a small
        # change in calculation between <py38 and >=py38
        for method in checker.LOG_METHODS:
            self._assert_has_errors(code.format(method), checker,
                                    expected_errors=[(4, mock.ANY, 'C310')])

        code = """
               import logging
               LOG = logging.getLogger()
               LOG.log(logging.DEBUG, "Volume %s caught fire and is at %d"
                       " degrees C and climbing.", ('volume1', 500))
               """
        # We don't assert on specific column numbers since there is a small
        # change in calculation between <py38 and >=py38
        self._assert_has_errors(code, checker,
                                expected_errors=[(4, mock.ANY, 'C310')])

    def test_opt_type_registration_args(self):
        checker = checks.CheckOptRegistrationArgs
        code = """
               CONF.register_opts([opt1, opt2, opt3])
               CONF.register_opts((opt4, opt5))
               CONF.register_opt(lonely_opt)
               CONF.register_opts([OPT1, OPT2], group="group_of_opts")
               CONF.register_opt(single_opt, group=blah)
               """
        self._assert_has_no_errors(code, checker)

        code = """
               CONF.register_opt([opt4, opt5, opt6])
               CONF.register_opt((opt7, opt8))
               CONF.register_opts(lonely_opt)
               CONF.register_opt((an_opt, another_opt))
               """
        # We don't assert on specific column numbers since there is a small
        # change in calculation between <py38 and >=py38
        self._assert_has_errors(code, checker,
                                expected_errors=[(1, 18, 'C311'),
                                                 (2, mock.ANY, 'C311'),
                                                 (3, mock.ANY, 'C311'),
                                                 (4, mock.ANY, 'C311')])

        code = """
               CONF.register_opt(single_opt)
               CONF.register_opts(other_opt)
               CONF.register_opt(multiple_opts)
               tuple_opts = (one_opt, two_opt)
               CONF.register_opts(tuple_opts)
               """
        self._assert_has_errors(code, checker,
                                expected_errors=[(2, 19, 'C311'),
                                                 (3, 18, 'C311')])

    def test_no_mutable_default_args(self):
        self.assertEqual(0, len(list(checks.no_mutable_default_args(
            "def foo (bar):"))))
        self.assertEqual(1, len(list(checks.no_mutable_default_args(
            "def foo (bar=[]):"))))
        self.assertEqual(1, len(list(checks.no_mutable_default_args(
            "def foo (bar={}):"))))

    def test_no_log_warn(self):
        code = """
                  LOG.warn("LOG.warn is deprecated")
               """
        errors = [(1, 0, 'C338')]
        self._assert_has_errors(code, checks.no_log_warn,
                                expected_errors=errors)
        code = """
                  LOG.warning("LOG.warn is deprecated")
               """
        self._assert_has_no_errors(code, checks.no_log_warn)

    def test_check_datetime_now(self):
        self.assertEqual(1, len(list(checks.check_datetime_now(
            "datetime.now", False))))
        self.assertEqual(0, len(list(checks.check_datetime_now(
            "timeutils.utcnow", False))))

    def test_check_datetime_now_noqa(self):
        self.assertEqual(0, len(list(checks.check_datetime_now(
                                     "datetime.now()  # noqa", True))))

    def test_check_timeutils_strtime(self):
        self.assertEqual(1, len(list(checks.check_timeutils_strtime(
            "timeutils.strtime"))))
        self.assertEqual(0, len(list(checks.check_timeutils_strtime(
            "strftime"))))

    def test_no_print_statements(self):
        self.assertEqual(0, len(list(checks.check_no_print_statements(
            "a line with no print statement",
            "cinder/file.py", False))))
        self.assertEqual(1, len(list(checks.check_no_print_statements(
            "print('My print statement')",
            "cinder/file.py", False))))
        self.assertEqual(0, len(list(checks.check_no_print_statements(
            "print('My print statement in cinder/cmd, which is ok.')",
            "cinder/cmd/file.py", False))))
        self.assertEqual(0, len(list(checks.check_no_print_statements(
            "print('My print statement that I just must have.')",
            "cinder.tests.unit/file.py", True))))
        self.assertEqual(1, len(list(checks.check_no_print_statements(
            "print ('My print with space')",
            "cinder/volume/anotherFile.py", False))))

    def test_dict_constructor_with_list_copy(self):
        self.assertEqual(1, len(list(checks.dict_constructor_with_list_copy(
            "    dict([(i, connect_info[i])"))))

        self.assertEqual(1, len(list(checks.dict_constructor_with_list_copy(
            "    attrs = dict([(k, _from_json(v))"))))

        self.assertEqual(1, len(list(checks.dict_constructor_with_list_copy(
            "        type_names = dict((value, key) for key, value in"))))

        self.assertEqual(1, len(list(checks.dict_constructor_with_list_copy(
            "   dict((value, key) for key, value in"))))

        self.assertEqual(1, len(list(checks.dict_constructor_with_list_copy(
            "foo(param=dict((k, v) for k, v in bar.items()))"))))

        self.assertEqual(1, len(list(checks.dict_constructor_with_list_copy(
            " dict([[i,i] for i in range(3)])"))))

        self.assertEqual(1, len(list(checks.dict_constructor_with_list_copy(
            "  dd = dict([i,i] for i in range(3))"))))

        self.assertEqual(0, len(list(checks.dict_constructor_with_list_copy(
            "  dict()"))))

        self.assertEqual(0, len(list(checks.dict_constructor_with_list_copy(
            "        create_kwargs = dict(snapshot=snapshot,"))))

        self.assertEqual(0, len(list(checks.dict_constructor_with_list_copy(
            "      self._render_dict(xml, data_el, data.__dict__)"))))

    def test_validate_assertTrue(self):
        test_value = True
        self.assertEqual(0, len(list(checks.validate_assertTrue(
            "assertTrue(True)", 'cinder/volume/stuff/a_file.py'))))
        self.assertEqual(0, len(list(checks.validate_assertTrue(
            "assertTrue(True)", 'cinder/tests/unit/test_file.py'))))
        self.assertEqual(0, len(list(checks.validate_assertTrue(
            "assertEqual(True, %s)" % test_value,
            'cinder/volume/stuff/a_file.py'))))
        self.assertEqual(1, len(list(checks.validate_assertTrue(
            "assertEqual(True, %s)" % test_value,
            'cinder/tests/unit/test_file.py'))))

    @ddt.unpack
    @ddt.data(
        (1, 'LOG.info', "cinder/tests/unit/fake.py", False),
        (1, 'LOG.warning', "cinder/tests/fake.py", False),
        (1, 'LOG.error', "cinder/tests/fake.py", False),
        (1, 'LOG.exception', "cinder/tests/fake.py", False),
        (1, 'LOG.debug', "cinder/tests/fake.py", False),
        (0, 'LOG.info.assert_called_once_with', "cinder/tests/fake.py", False),
        (0, 'some.LOG.error.call', "cinder/tests/fake.py", False),
        (0, 'LOG.warning', "cinder/tests/unit/fake.py", True))
    def test_no_test_log(self, first, second, third, fourth):
        self.assertEqual(first, len(list(checks.no_test_log(
            "%s('arg')" % second, third, fourth))))

    @ddt.unpack
    @ddt.data(
        (1, 'import mock'),
        (0, 'from unittest import mock'),
        (1, 'from mock import patch'),
        (0, 'from unittest.mock import patch'))
    def test_no_third_party_mock(self, err_count, line):
        self.assertEqual(err_count, len(list(checks.no_third_party_mock(
            line))))



# Copyright (C) 2020, 2021, Hitachi, Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""REST interface fibre channel module for Hitachi HBSD Driver."""

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils

from cinder import exception
from cinder.volume import configuration
from cinder.volume.drivers.hitachi import hbsd_rest as rest
from cinder.volume.drivers.hitachi import hbsd_rest_api as rest_api
from cinder.volume.drivers.hitachi import hbsd_utils as utils
from cinder.zonemanager import utils as fczm_utils

FC_VOLUME_OPTS = [
    cfg.BoolOpt(
        'hitachi_zoning_request',
        default=False,
        help='If True, the driver will configure FC zoning between the server '
             'and the storage system provided that FC zoning manager is '
             'enabled.'),
]

_FC_HMO_DISABLE_IO = 91

_MSG_EXCEED_HOST_GROUP_MAX = "could not find empty Host group ID for adding."

LOG = logging.getLogger(__name__)
MSG = utils.HBSDMsg

CONF = cfg.CONF
CONF.register_opts(FC_VOLUME_OPTS, group=configuration.SHARED_CONF_GROUP)


class HBSDRESTFC(rest.HBSDREST):
    """REST interface fibre channel class for Hitachi HBSD Driver."""

    def __init__(self, conf, storage_protocol, db):
        """Initialize instance variables."""
        super(HBSDRESTFC, self).__init__(conf, storage_protocol, db)
        self._lookup_service = fczm_utils.create_lookup_service()

    def connect_storage(self):
        """Prepare for using the storage."""
        target_ports = self.conf.hitachi_target_ports
        compute_target_ports = self.conf.hitachi_compute_target_ports
        available_ports = []
        available_compute_ports = []

        super(HBSDRESTFC, self).connect_storage()
        # The port attributes must contain TAR.
        params = {'portAttributes': 'TAR'}
        port_list = self.client.get_ports(params=params)
        for port in set(target_ports + compute_target_ports):
            if port not in [port_data['portId'] for port_data in port_list]:
                utils.output_log(MSG.INVALID_PORT, port=port,
                                 additional_info='portAttributes: not TAR')
        for port_data in port_list:
            port = port_data['portId']
            if port not in set(target_ports + compute_target_ports):
                continue
            secure_fc_port = True
            can_port_schedule = True
            if hasattr(
                    self.conf,
                    self.driver_info['param_prefix'] + '_port_scheduler'):
                port_scheduler_param = self.conf.hitachi_port_scheduler
            else:
                port_scheduler_param = False
            if (port_data['portType'] not in ['FIBRE', 'FCoE'] or
                    not port_data['lunSecuritySetting']):
                secure_fc_port = False
            elif (port in set(target_ports + compute_target_ports) and
                  port_scheduler_param and not (
                      port_data.get('fabricMode') and
                      port_data.get('portConnection') == 'PtoP')):
                can_port_schedule = False
            if not secure_fc_port or not can_port_schedule:
                utils.output_log(
                    MSG.INVALID_PORT, port=port,
                    additional_info='portType: %s, lunSecuritySetting: %s, '
                    'fabricMode: %s, portConnection: %s' %
                    (port_data['portType'],
                     port_data.get('lunSecuritySetting'),
                     port_data.get('fabricMode'),
                     port_data.get('portConnection')))
            if not secure_fc_port:
                continue
            wwn = port_data.get('wwn')
            if target_ports and port in target_ports and can_port_schedule:
                available_ports.append(port)
                self.storage_info['wwns'][port] = wwn
            if (compute_target_ports and port in compute_target_ports and
                    can_port_schedule):
                available_compute_ports.append(port)
                self.storage_info['wwns'][port] = wwn

        if target_ports:
            for port in target_ports:
                if port in available_ports:
                    self.storage_info['controller_ports'].append(port)
        if compute_target_ports:
            for port in compute_target_ports:
                if port in available_compute_ports:
                    self.storage_info['compute_ports'].append(port)

        self.check_ports_info()
        utils.output_log(MSG.SET_CONFIG_VALUE, object='port-wwn list',
                         value=self.storage_info['wwns'])

    def check_param(self):
        """Check parameter values and consistency among them."""
        super(HBSDRESTFC, self).check_param()
        self.check_opts(self.conf, FC_VOLUME_OPTS)

    def create_target_to_storage(self, port, connector, hba_ids):
        """Create a host group on the specified port."""
        wwpns = self.get_hba_ids_from_connector(connector)
        target_name = '%(prefix)s-%(wwpns)s' % {
            'prefix': self.driver_info['driver_prefix'],
            'wwpns': min(wwpns),
        }
        try:
            body = {'portId': port,
                    'hostGroupName': target_name}
            gid = self.client.add_host_grp(body, no_log=True)
        except Exception:
            params = {'portId': port}
            host_grp_list = self.client.get_host_grps(params)
            for host_grp_data in host_grp_list:
                if host_grp_data['hostGroupName'] == target_name:
                    return target_name, host_grp_data['hostGroupNumber']
            raise
        return target_name, gid

    def set_hba_ids(self, port, gid, hba_ids):
        """Connect all specified HBAs with the specified port."""
        registered_wwns = []
        for wwn in hba_ids:
            try:
                self.client.add_hba_wwn(port, gid, wwn, no_log=True)
                registered_wwns.append(wwn)
            except exception.VolumeDriverException as ex:
                utils.output_log(MSG.ADD_HBA_WWN_FAILED, port=port, gid=gid,
                                 wwn=wwn)
                if (self.get_port_scheduler_param() and
                        utils.safe_get_err_code(ex.kwargs.get('errobj'))
                        == rest_api.EXCEED_WWN_MAX):
                    raise ex
        if not registered_wwns:
            msg = utils.output_log(MSG.NO_HBA_WWN_ADDED_TO_HOST_GRP, port=port,
                                   gid=gid)
            self.raise_error(msg)

    def set_target_mode(self, port, gid):
        """Configure the host group to meet the environment."""
        body = {'hostMode': 'LINUX/IRIX'}
        if self.conf.hitachi_rest_disable_io_wait:
            body['hostModeOptions'] = [_FC_HMO_DISABLE_IO]
        if self.conf.hitachi_host_mode_options:
            if 'hostModeOptions' not in body:
                body['hostModeOptions'] = []
            for opt in self.conf.hitachi_host_mode_options:
                if int(opt) not in body['hostModeOptions']:
                    body['hostModeOptions'].append(int(opt))
        self.client.modify_host_grp(port, gid, body, ignore_all_errors=True)

    def _get_hwwns_in_hostgroup(self, port, gid, wwpns):
        """Return WWN registered with the host group."""
        hwwns_in_hostgroup = []
        for hba_wwn in self.client.get_hba_wwns(port, gid):
            hwwn = hba_wwn['hostWwn']
            if hwwn in wwpns:
                hwwns_in_hostgroup.append(hwwn)
        return hwwns_in_hostgroup

    def _set_target_info(self, targets, host_grps, wwpns):
        """Set the information of the host group having the specified WWN."""
        for host_grp in host_grps:
            port = host_grp['portId']
            gid = host_grp['hostGroupNumber']
            hwwns_in_hostgroup = self._get_hwwns_in_hostgroup(port, gid, wwpns)
            if hwwns_in_hostgroup:
                targets['info'][port] = True
                targets['list'].append((port, gid))
                LOG.debug(
                    'Found wwpns in host group. (port: %(port)s, '
                    'gid: %(gid)s, wwpns: %(wwpns)s)',
                    {'port': port, 'gid': gid, 'wwpns': hwwns_in_hostgroup})
                return True
        return False

    def _get_hwwns_in_hostgroup_by_name(self, port, host_group_name, wwpns):
        """Return WWN registered with the host group of the specified name."""
        hba_wwns = self.client.get_hba_wwns_by_name(port, host_group_name)
        return [hba_wwn for hba_wwn in hba_wwns if hba_wwn['hostWwn'] in wwpns]

    def _set_target_info_by_names(self, targets, port, target_names, wwpns):
        """Set the information of the host group having the specified name and

        the specified WWN.
        """
        for target_name in target_names:
            hwwns_in_hostgroup = self._get_hwwns_in_hostgroup_by_name(
                port, target_name, wwpns)
            if hwwns_in_hostgroup:
                gid = hwwns_in_hostgroup[0]['hostGroupNumber']
                targets['info'][port] = True
                targets['list'].append((port, gid))
                LOG.debug(
                    'Found wwpns in host group. (port: %(port)s, '
                    'gid: %(gid)s, wwpns: %(wwpns)s)',
                    {'port': port, 'gid': gid, 'wwpns':
                     [hwwn['hostWwn'] for hwwn in hwwns_in_hostgroup]})
                return True
        return False

    def find_targets_from_storage(
            self, targets, connector, target_ports):
        """Find mapped ports, memorize them and return unmapped port count."""
        wwpns = self.get_hba_ids_from_connector(connector)
        target_names = [
            '%(prefix)s-%(wwpns)s' % {
                'prefix': self.driver_info['driver_prefix'],
                'wwpns': min(wwpns),
            }
        ]
        if 'ip' in connector:
            target_names.append(
                '%(prefix)s-%(ip)s' % {
                    'prefix': self.driver_info['driver_prefix'],
                    'ip': connector['ip'],
                }
            )
        not_found_count = 0
        for port in target_ports:
            targets['info'][port] = False
            if self._set_target_info_by_names(
                    targets, port, target_names, wwpns):
                continue
            host_grps = self.client.get_host_grps({'portId': port})
            if self._set_target_info(
                targets, [hg for hg in host_grps if hg['hostGroupName'] not in
                          target_names], wwpns):
                pass
            else:
                not_found_count += 1

        if self.get_port_scheduler_param():
            """
            When port scheduler feature is enabled,
            it is OK to find any mapped port. so:
            - return 0, if any mapped port is found
            - return port count, if no mapped port is found.
            It is no case with both not_found_count and len(target_ports) are
            zero, bcz it must be failed in param checker if any target ports
            are not defined.
            """
            return (not_found_count if not_found_count == len(target_ports)
                    else 0)

        return not_found_count

    def initialize_connection(self, volume, connector, is_snapshot=False):
        """Initialize connection between the server and the volume."""
        conn_info, map_info = super(HBSDRESTFC, self).initialize_connection(
            volume, connector, is_snapshot)
        if self.conf.hitachi_zoning_request:
            if (self.get_port_scheduler_param() and
                    not self.is_controller(connector)):
                init_targ_map = map_info
            else:
                init_targ_map = utils.build_initiator_target_map(
                    connector, conn_info['data']['target_wwn'],
                    self._lookup_service)
            if init_targ_map:
                conn_info['data']['initiator_target_map'] = init_targ_map
            fczm_utils.add_fc_zone(conn_info)
        return conn_info

    def terminate_connection(self, volume, connector):
        """Terminate connection between the server and the volume."""
        conn_info = super(HBSDRESTFC, self).terminate_connection(
            volume, connector)
        if self.conf.hitachi_zoning_request:
            if conn_info and conn_info['data']['target_wwn']:
                init_targ_map = utils.build_initiator_target_map(
                    connector, conn_info['data']['target_wwn'],
                    self._lookup_service)
                if init_targ_map:
                    conn_info['data']['initiator_target_map'] = init_targ_map
            fczm_utils.remove_fc_zone(conn_info)
        return conn_info

    def _get_wwpns(self, port, hostgroup):
        """Get WWPN from a port and the host group."""
        wwpns = []
        hba_wwns = self.client.get_hba_wwns_by_name(port, hostgroup)
        for hba_wwn in hba_wwns:
            wwpns.append(hba_wwn['hostWwn'])
        return wwpns

    def set_terminate_target(self, fake_connector, port_hostgroup_map):
        """Set necessary information in connector in terminate."""
        wwpns = set()
        for port, hostgroups in port_hostgroup_map.items():
            for hostgroup in hostgroups:
                wwpns.update(self._get_wwpns(port, hostgroup))
        fake_connector['wwpns'] = list(wwpns)

    def set_device_map(self, targets, hba_ids, volume):
        active_hba_ids = []
        target_wwns = []
        active_target_wwns = []
        vol_id = volume['id'] if volume and 'id' in volume.keys() else ""

        if not self.get_port_scheduler_param():
            return None, hba_ids

        for port in targets['info'].keys():
            target_wwns.append(self.storage_info['wwns'][port])

        devmap = self._lookup_service.get_device_mapping_from_network(
            hba_ids, target_wwns)

        for fabric_name in devmap.keys():
            active_hba_ids.extend(
                devmap[fabric_name]['initiator_port_wwn_list'])
            active_target_wwns.extend(
                devmap[fabric_name]['target_port_wwn_list'])

        active_hba_ids = list(set(active_hba_ids))
        if not active_hba_ids:
            msg = utils.output_log(MSG.NO_ACTIVE_WWN, wwn=', '.join(hba_ids),
                                   volume=vol_id)
            self.raise_error(msg)

        active_target_wwns = list(set(active_target_wwns))
        if not active_target_wwns:
            port_wwns = ""
            for port in targets['info'].keys():
                if port_wwns:
                    port_wwns += ", "
                port_wwns += ("port, WWN: " + port +
                              ", " + self.storage_info['wwns'][port])
            msg = utils.output_log(
                MSG.NO_PORT_WITH_ACTIVE_WWN, port_wwns=port_wwns,
                volume=vol_id)
            self.raise_error(msg)

        return devmap, active_hba_ids

    def build_wwpn_groups(self, wwpns, connector):
        count = 1
        return ([wwpns[i:i + count] for i in range(0, len(wwpns), count)])

    def _create_target_to_any_port(
            self, targets, ports, connector, hba_ids, fabric_name):
        for port in ports:
            index = self.get_port_index_to_be_used(ports, fabric_name)
            try:
                self.create_target(
                    targets, ports[index], connector, hba_ids)
                return
            except exception.VolumeDriverException as ex:
                if ((utils.safe_get_message_id(ex.kwargs.get('errobj'))
                        == rest_api.MSGID_SPECIFIED_OBJECT_DOES_NOT_EXIST)
                    or (_MSG_EXCEED_HOST_GROUP_MAX
                        in utils.safe_get_message(ex.kwargs.get('errobj')))):
                    utils.output_log(
                        MSG.HOST_GROUP_NUMBER_IS_MAXIMUM, port=ports[index])
                elif (utils.safe_get_err_code(ex.kwargs.get('errobj'))
                        == rest_api.EXCEED_WWN_MAX):
                    utils.output_log(
                        MSG.WWN_NUMBER_IS_MAXIMUM, port=ports[index],
                        wwn=", ". join(hba_ids))
                else:
                    raise ex

        msg = utils.output_log(
            MSG.HOST_GROUP_OR_WWN_IS_NOT_AVAILABLE, ports=', '.join(ports))
        self.raise_error(msg)

    def create_target_by_port_scheduler(
            self, devmap, targets, connector, volume):
        available_ports = []
        active_ports = []

        if not devmap:
            msg = utils.output_log(MSG.ZONE_MANAGER_IS_NOT_AVAILABLE)
            self.raise_error(msg)
        for fabric_name in devmap.keys():
            available_ports = []
            active_ports = []
            active_initiator_wwns = devmap[
                fabric_name]['initiator_port_wwn_list']
            wwpn_groups = self.build_wwpn_groups(
                active_initiator_wwns, connector)
            for port, wwn in self.storage_info['wwns'].items():
                if wwn in devmap[fabric_name]['target_port_wwn_list']:
                    available_ports.append(port)
            target_ports = self.get_target_ports(connector)
            filter_ports = self.filter_target_ports(target_ports, volume)
            for port in target_ports:
                if port in available_ports and port in filter_ports:
                    active_ports.append(port)
                elif port not in available_ports and port in filter_ports:
                    utils.output_log(
                        MSG.INVALID_PORT_BY_ZONE_MANAGER, port=port)
            for wwpns in wwpn_groups:
                try:
                    self._create_target_to_any_port(
                        targets, active_ports, connector, wwpns, fabric_name)
                except exception.VolumeDriverException:
                    with excutils.save_and_reraise_exception():
                        self.clean_mapping_targets(targets)

    def set_target_map_info(self, targets, hba_ids, port):
        for hba_id in hba_ids:
            target_map = {hba_id: [self.storage_info['wwns'][port]]}
            targets['target_map'].update(target_map)



# Copyright (c) 2019 Zadara Storage, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Volume driver for Zadara Virtual Private Storage Array (VPSA).

This driver requires VPSA with API version 15.07 or higher.
"""

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import strutils
import six

from cinder.common import constants
from cinder import exception as cinder_exception
from cinder.i18n import _
from cinder import interface
from cinder.objects import fields
from cinder.volume import configuration
from cinder.volume import driver
from cinder.volume.drivers.zadara import common
from cinder.volume.drivers.zadara import exception as zadara_exception
from cinder.volume import volume_utils

CONF = cfg.CONF
CONF.register_opts(common.zadara_opts, group=configuration.SHARED_CONF_GROUP)

LOG = logging.getLogger(__name__)

cinder_opts = [
    cfg.BoolOpt('zadara_use_iser',
                default=True,
                help='VPSA - Use ISER instead of iSCSI'),
    cfg.StrOpt('zadara_vol_name_template',
               default='OS_%s',
               help='VPSA - Default template for VPSA volume names')]


@interface.volumedriver
class ZadaraVPSAISCSIDriver(driver.ISCSIDriver):
    """Zadara VPSA iSCSI/iSER volume driver.

    .. code-block:: none

      Version history:
        15.07 - Initial driver
        16.05 - Move from httplib to requests
        19.08 - Add API access key authentication option
        20.01 - Move to json format from xml. Provide manage/unmanage
                volume/snapshot feature
        20.12-01 - Merging with the common code for all the openstack drivers
        20.12-02 - Common code changed as part of fixing
                   Zadara github issue #18723
        20.12-03 - Adding the metadata support while creating volume to
                   configure vpsa.
        20.12-20 - IPv6 connectivity support for Cinder driver
        20.12-24 - Optimizing get manageable volumes and snapshots
    """

    VERSION = '20.12-24'

    # ThirdPartySystems wiki page
    CI_WIKI_NAME = "ZadaraStorage_VPSA_CI"

    def __init__(self, *args, **kwargs):
        super(ZadaraVPSAISCSIDriver, self).__init__(*args, **kwargs)
        self.vpsa = None
        self.configuration.append_config_values(common.zadara_opts)
        self.configuration.append_config_values(cinder_opts)
        # The valid list of volume options that can be specified
        # as the metadata while creating cinder volume
        self.vol_options = ['crypt', 'compress',
                            'dedupe', 'attachpolicies']

    @staticmethod
    def get_driver_options():
        driver_opts = []
        driver_opts.extend(common.zadara_opts)
        driver_opts.extend(cinder_opts)
        return driver_opts

    def _check_access_key_validity(self):
        try:
            self.vpsa._check_access_key_validity()
        except common.exception.ZadaraInvalidAccessKey:
            raise zadara_exception.ZadaraCinderInvalidAccessKey()

    def do_setup(self, context):
        """Any initialization the volume driver does while starting.

        Establishes initial connection with VPSA and retrieves access_key.
        Need to pass driver_ssl_cert_path here (and not fetch it from the
        config opts directly in common code), because this config option is
        different for different drivers and so cannot be figured in the
        common code.
        """
        driver_ssl_cert_path = self.configuration.driver_ssl_cert_path
        self.vpsa = common.ZadaraVPSAConnection(self.configuration,
                                                driver_ssl_cert_path, True)
        self._check_access_key_validity()

    def check_for_setup_error(self):
        """Returns an error (exception) if prerequisites aren't met."""
        self._check_access_key_validity()

    def local_path(self, volume):
        """Return local path to existing local volume."""
        raise NotImplementedError()

    def _get_zadara_vol_template_name(self, vol_name):
        return self.configuration.zadara_vol_name_template % vol_name

    def _get_vpsa_volume(self, volume, raise_exception=True):
        vpsa_volume = None
        if volume.provider_location:
            vpsa_volume = (self.vpsa._get_vpsa_volume_by_id(
                           volume.provider_location))
        else:
            vol_name = self._get_zadara_vol_template_name(volume.name)
            vpsa_volume = self.vpsa._get_vpsa_volume(vol_name)

        if not vpsa_volume:
            vol_name = self._get_zadara_vol_template_name(volume.name)
            msg = (_('Backend Volume %(name)s not found') % {'name': vol_name})
            if raise_exception:
                LOG.error(msg)
                raise cinder_exception.VolumeDriverException(message=msg)
            LOG.warning(msg)
        return vpsa_volume

    def vpsa_send_cmd(self, cmd, **kwargs):
        try:
            response = self.vpsa.send_cmd(cmd, **kwargs)
        except common.exception.UnknownCmd as e:
            raise cinder_exception.UnknownCmd(cmd=e.cmd)
        except common.exception.SessionRequestException as e:
            raise zadara_exception.ZadaraSessionRequestException(msg=e.msg)
        except common.exception.BadHTTPResponseStatus as e:
            raise cinder_exception.BadHTTPResponseStatus(status=e.status)
        except common.exception.FailedCmdWithDump as e:
            raise cinder_exception.FailedCmdWithDump(status=e.status,
                                                     data=e.data)
        except common.exception.ZadaraInvalidAccessKey:
            raise zadara_exception.ZadaraCinderInvalidAccessKey()
        return response

    def _validate_existing_ref(self, existing_ref):
        """Validates existing ref"""
        if not existing_ref.get('name'):
            raise cinder_exception.ManageExistingInvalidReference(
                existing_ref=existing_ref,
                reason=_("manage_existing requires a 'name'"
                         " key to identify an existing volume."))

    def _get_volume_metadata(self, volume):
        if 'metadata' in volume:
            return volume.metadata
        if 'volume_metadata' in volume:
            metadata = volume.volume_metadata
            return {m['key']: m['value'] for m in metadata}
        return {}

    def is_valid_metadata(self, metadata):
        LOG.debug('Metadata while creating volume: %(metadata)s',
                  {'metadata': metadata})
        # Check the values allowed for provided metadata
        return all(value in ('YES', 'NO')
                   for key, value in metadata.items()
                   if key in self.vol_options)

    def create_volume(self, volume):
        """Create volume."""
        vol_name = self._get_zadara_vol_template_name(volume.name)

        # Collect the volume metadata if any provided and validate it
        metadata = self._get_volume_metadata(volume)
        if not self.is_valid_metadata(metadata):
            msg = (_('Invalid metadata for Volume %s') % vol_name)
            LOG.error(msg)
            raise cinder_exception.VolumeDriverException(message=msg)

        data = self.vpsa_send_cmd('create_volume',
                                  name=vol_name,
                                  size=volume.size,
                                  metadata=metadata)

        return {'provider_location': data.get('vol_name')}

    def delete_volume(self, volume):
        """Delete volume.

        Return ok if doesn't exist. Auto detach from all servers.
        """
        vpsa_volume = self._get_vpsa_volume(volume, False)
        if not vpsa_volume:
            return

        self.vpsa._detach_vpsa_volume(vpsa_vol=vpsa_volume)

        # Delete volume
        self.vpsa_send_cmd('delete_volume', vpsa_vol=vpsa_volume['name'])

    def create_snapshot(self, snapshot):
        """Creates a snapshot."""

        LOG.debug('Create snapshot: %s', snapshot.name)

        vpsa_volume = self._get_vpsa_volume(snapshot.volume)
        # Retrieve the CG name for the base volume
        cg_name = vpsa_volume['cg_name']
        data = self.vpsa_send_cmd('create_snapshot',
                                  cg_name=cg_name,
                                  snap_name=snapshot.name)

        return {'provider_location': data.get('snapshot_name')}

    def delete_snapshot(self, snapshot):
        """Deletes a snapshot."""

        LOG.debug('Delete snapshot: %s', snapshot.name)

        vpsa_volume = self._get_vpsa_volume(snapshot.volume, False)
        if not vpsa_volume:
            # If the volume isn't present, then don't attempt to delete
            return

        # Retrieve the CG name for the base volume
        cg_name = vpsa_volume['cg_name']
        snap_id = self.vpsa._get_snap_id(cg_name, snapshot.name)
        if not snap_id:
            # If the snapshot isn't present, then don't attempt to delete
            LOG.warning('snapshot: snapshot %s not found, '
                        'skipping delete operation', snapshot.name)
            return

        self.vpsa_send_cmd('delete_snapshot',
                           snap_id=snap_id)

    def create_volume_from_snapshot(self, volume, snapshot):
        """Creates a volume from a snapshot."""

        LOG.debug('Creating volume from snapshot: %s', snapshot.name)

        vpsa_volume = self._get_vpsa_volume(snapshot.volume, False)
        if not vpsa_volume:
            LOG.error('Snapshot %(name)s not found.',
                      {'name': snapshot.name})
            raise cinder_exception.SnapshotNotFound(snapshot_id=snapshot.id)

        # Retrieve the CG name for the base volume
        cg_name = vpsa_volume['cg_name']
        snap_id = self.vpsa._get_snap_id(cg_name, snapshot.name)
        if not snap_id:
            LOG.error('Snapshot %(name)s not found',
                      {'name': snapshot.name})
            raise cinder_exception.SnapshotNotFound(snapshot_id=snapshot.id)

        volume_name = self._get_zadara_vol_template_name(volume.name)
        self.vpsa_send_cmd('create_clone_from_snap',
                           cg_name=cg_name,
                           name=volume_name,
                           snap_id=snap_id)

        vpsa_volume = self._get_vpsa_volume(volume)
        if volume.size > snapshot.volume_size:
            self.extend_volume(volume, volume.size)
        return {'provider_location': vpsa_volume.get('name')}

    def create_cloned_volume(self, volume, src_vref):
        """Creates a clone of the specified volume."""

        LOG.debug('Creating clone of volume: %s', src_vref.name)

        vpsa_volume = self._get_vpsa_volume(src_vref)
        # Retrieve the CG name for the base volume
        cg_name = vpsa_volume['cg_name']
        volume_name = self._get_zadara_vol_template_name(volume.name)
        self.vpsa_send_cmd('create_clone',
                           cg_name=cg_name,
                           name=volume_name)

        vpsa_volume = self._get_vpsa_volume(volume)
        if volume.size > src_vref.size:
            self.extend_volume(volume, volume.size)
        return {'provider_location': vpsa_volume.get('name')}

    def extend_volume(self, volume, new_size):
        """Extend an existing volume."""
        # Get volume
        vpsa_volume = self._get_vpsa_volume(volume)
        size = vpsa_volume['virtual_capacity']
        if new_size < size:
            raise cinder_exception.InvalidInput(
                reason=_('%(new_size)s < current size %(size)s') %
                {'new_size': new_size, 'size': size})

        expand_size = new_size - size
        self.vpsa_send_cmd('expand_volume',
                           vpsa_vol=vpsa_volume['name'],
                           size=expand_size)

    def create_export(self, context, volume, vg=None):
        """Irrelevant for VPSA volumes. Export created during attachment."""
        pass

    def ensure_export(self, context, volume):
        """Irrelevant for VPSA volumes. Export created during attachment."""
        pass

    def remove_export(self, context, volume):
        """Irrelevant for VPSA volumes. Export removed during detach."""
        pass

    def get_manageable_volumes(self, cinder_volumes, marker, limit, offset,
                               sort_keys, sort_dirs):
        """List volumes on the backend available for management by Cinder"""
        # Get all vpsa volumes
        all_vpsa_volumes = self.vpsa._get_all_vpsa_volumes()

        # Create a dictionary of existing volumes
        existing_vols = {}
        for cinder_vol in cinder_volumes:
            if cinder_vol.provider_location:
                volumes = (list(filter(lambda volume:
                           (volume['name'] == cinder_vol.provider_location),
                           all_vpsa_volumes)))
            else:
                cinder_name = (self._get_zadara_vol_template_name(
                               cinder_vol.name))
                volumes = (list(filter(lambda volume:
                                (volume['display_name'] == cinder_name),
                                all_vpsa_volumes)))
            for volume in volumes:
                existing_vols[volume['name']] = cinder_vol.id

        # Filter out all volumes already attached to any server
        volumes_in_use = {}
        volumes_not_available = {}
        for volume in all_vpsa_volumes:
            if volume['name'] in existing_vols:
                continue

            if volume['status'] == 'In-use':
                volumes_in_use[volume['name']] =\
                    self.vpsa._get_servers_attached_to_volume(volume)
                continue

            if volume['status'] != 'Available':
                volumes_not_available[volume['name']] = volume['display_name']
                continue

        manageable_vols = []
        for vpsa_volume in all_vpsa_volumes:
            vol_name = vpsa_volume['name']
            vol_display_name = vpsa_volume['display_name']
            cinder_id = existing_vols.get(vol_name)
            not_safe_msgs = []

            if vol_name in volumes_in_use:
                host_list = volumes_in_use[vol_name]
                not_safe_msgs.append(_('Volume connected to host(s) %s')
                                     % host_list)

            elif vol_name in volumes_not_available:
                not_safe_msgs.append(_('Volume not available'))

            if cinder_id:
                not_safe_msgs.append(_('Volume already managed'))

            is_safe = (len(not_safe_msgs) == 0)
            reason_not_safe = ' && '.join(not_safe_msgs)

            manageable_vols.append({
                'reference': {'name': vol_display_name},
                'size': vpsa_volume['virtual_capacity'],
                'safe_to_manage': is_safe,
                'reason_not_safe': reason_not_safe,
                'cinder_id': cinder_id,
            })

        return volume_utils.paginate_entries_list(
            manageable_vols, marker, limit, offset, sort_keys, sort_dirs)

    def manage_existing(self, volume, existing_ref):
        """Bring an existing volume into cinder management"""
        self._validate_existing_ref(existing_ref)

        # Check if the volume exists in vpsa
        name = existing_ref['name']
        vpsa_volume = self.vpsa._get_vpsa_volume(name)
        if not vpsa_volume:
            msg = (_('Volume %(name)s could not be found. '
                     'It might be already deleted') % {'name': name})
            LOG.error(msg)
            raise cinder_exception.ManageExistingInvalidReference(
                existing_ref=existing_ref,
                reason=msg)

        # Check if the volume is available
        if vpsa_volume['status'] != 'Available':
            msg = (_('Existing volume %(name)s is not available')
                   % {'name': name})
            LOG.error(msg)
            raise cinder_exception.ManageExistingInvalidReference(
                existing_ref=existing_ref,
                reason=msg)

        # Rename the volume to cinder specified name
        new_name = self._get_zadara_vol_template_name(volume.name)
        new_vpsa_volume = self.vpsa._get_vpsa_volume(new_name)
        if new_vpsa_volume:
            msg = (_('Volume %(new_name)s already exists')
                   % {'new_name': new_name})
            LOG.error(msg)
            raise cinder_exception.VolumeDriverException(message=msg)

        data = self.vpsa_send_cmd('rename_volume',
                                  vpsa_vol=vpsa_volume['name'],
                                  new_name=new_name)
        return {'provider_location': data.get('vol_name')}

    def manage_existing_get_size(self, volume, existing_ref):
        """Return size of volume to be managed by manage_existing"""
        # Check if the volume exists in vpsa
        self._validate_existing_ref(existing_ref)
        name = existing_ref['name']
        vpsa_volume = self.vpsa._get_vpsa_volume(name)
        if not vpsa_volume:
            msg = (_('Volume %(name)s could not be found. '
                     'It might be already deleted') % {'name': volume.name})
            LOG.error(msg)
            raise cinder_exception.ManageExistingInvalidReference(
                existing_ref=existing_ref,
                reason=msg)

        # Return the size of the volume
        return vpsa_volume['virtual_capacity']

    def unmanage(self, volume):
        """Removes the specified volume from Cinder management"""
        pass

    def get_manageable_snapshots(self, cinder_snapshots, marker, limit, offset,
                                 sort_keys, sort_dirs):
        """Interface to support listing manageable snapshots and volumes"""
        # Get all snapshots
        vpsa_snapshots = self.vpsa._get_all_vpsa_snapshots()

        # Get all snapshots of all volumes
        all_vpsa_snapshots = []
        for vpsa_snap in vpsa_snapshots:
            if (vpsa_snap['pool_name'] ==
                    self.configuration.zadara_vpsa_poolname):
                vpsa_snap['volume_name'] = vpsa_snap['volume_display_name']
                vpsa_snap['size'] = float(vpsa_snap['volume_capacity_mb'] /
                                          1024)
                all_vpsa_snapshots.append(vpsa_snap)

        existing_snapshots = {}
        for cinder_snapshot in cinder_snapshots:
            if cinder_snapshot.provider_location:
                snapshots = (list(filter(lambda snapshot:
                             ((snapshot['volume_ext_name'] ==
                               cinder_snapshot.volume.provider_location) and
                              (snapshot['name'] ==
                               cinder_snapshot.provider_location)),
                             all_vpsa_snapshots)))
            else:
                volume_name = (self._get_zadara_vol_template_name(
                    cinder_snapshot.volume_name))
                snapshots = (list(filter(lambda snapshot:
                             ((snapshot['volume_display_name'] ==
                               volume_name) and
                              (snapshot['display_name'] ==
                               cinder_snapshot.name)),
                             all_vpsa_snapshots)))
            for snapshot in snapshots:
                existing_snapshots[snapshot['name']] = cinder_snapshot.id

        manageable_snapshots = []
        try:
            unique_snapshots = []
            for snapshot in all_vpsa_snapshots:
                snap_id = snapshot['name']
                if snap_id in unique_snapshots:
                    continue

                cinder_id = existing_snapshots.get(snap_id)
                is_safe = True
                reason_not_safe = None

                if cinder_id:
                    is_safe = False
                    reason_not_safe = _("Snapshot already managed.")

                manageable_snapshots.append({
                    'reference': {'name': snapshot['display_name']},
                    'size': snapshot['size'],
                    'safe_to_manage': is_safe,
                    'reason_not_safe': reason_not_safe,
                    'cinder_id': cinder_id,
                    'extra_info': None,
                    'source_reference': {'name': snapshot['volume_name']},
                })

                unique_snapshots.append(snap_id)
            return volume_utils.paginate_entries_list(
                manageable_snapshots, marker, limit, offset,
                sort_keys, sort_dirs)
        except Exception as e:
            msg = (_('Exception: %s') % six.text_type(e))
            LOG.error(msg)
            raise

    def manage_existing_snapshot(self, snapshot, existing_ref):
        """Brings an existing backend storage object under Cinder management"""
        self._validate_existing_ref(existing_ref)

        snap_name = existing_ref['name']
        volume = self._get_vpsa_volume(snapshot.volume, False)
        if not volume:
            msg = (_('Source volume of snapshot %s could not be found.'
                     ' Invalid data') % snap_name)
            LOG.error(msg)
            raise cinder_exception.ManageExistingInvalidReference(
                existing_ref=existing_ref,
                reason=msg)

        # Check if the snapshot exists
        snap_id = self.vpsa._get_snap_id(volume['cg_name'], snap_name)
        if not snap_id:
            msg = (_('Snapshot %s could not be found. It might be'
                     ' already deleted') % snap_name)
            LOG.error(msg)
            raise cinder_exception.ManageExistingInvalidReference(
                existing_ref=existing_ref,
                reason=msg)

        new_name = snapshot.name
        new_snap_id = self.vpsa._get_snap_id(volume['cg_name'], new_name)
        if new_snap_id:
            msg = (_('Snapshot with name %s already exists') % new_name)
            LOG.debug(msg)
            return

        data = self.vpsa_send_cmd('rename_snapshot',
                                  snap_id=snap_id,
                                  new_name=new_name)
        return {'provider_location': data.get('snapshot_name')}

    def manage_existing_snapshot_get_size(self, snapshot, existing_ref):
        """Return size of snapshot to be managed by manage_existing"""
        # We do not have any size field for a snapshot.
        # We only have it on volumes. So, here just figure
        # out the parent volume of this snapshot and return its size
        self._validate_existing_ref(existing_ref)
        snap_name = existing_ref['name']
        volume = self._get_vpsa_volume(snapshot.volume, False)
        if not volume:
            msg = (_('Source volume of snapshot %s could not be found.'
                     ' Invalid data') % snap_name)
            LOG.error(msg)
            raise cinder_exception.ManageExistingInvalidReference(
                existing_ref=existing_ref,
                reason=msg)

        snap_id = self.vpsa._get_snap_id(volume['cg_name'], snap_name)
        if not snap_id:
            msg = (_('Snapshot %s could not be found. It might be '
                     'already deleted') % snap_name)
            LOG.error(msg)
            raise cinder_exception.ManageExistingInvalidReference(
                existing_ref=existing_ref,
                reason=msg)

        return volume['virtual_capacity']

    def unmanage_snapshot(self, snapshot):
        """Removes the specified snapshot from Cinder management"""
        pass

    def initialize_connection(self, volume, connector):
        """Attach volume to initiator/host.

        During this call VPSA exposes volume to particular Initiator. It also
        creates a 'server' entity for Initiator (if it was not created before)
        All necessary connection information is returned, including auth data.
        Connection data (target, LUN) is not stored in the DB.
        """
        # First: Check Active controller: if not valid, raise exception
        ctrl = self.vpsa._get_active_controller_details()
        if not ctrl:
            raise zadara_exception.ZadaraVPSANoActiveController()

        # Get/Create server name for IQN
        initiator_name = connector['initiator']
        vpsa_srv = self.vpsa._create_vpsa_server(iqn=initiator_name)
        if not vpsa_srv:
            raise zadara_exception.ZadaraServerCreateFailure(
                name=initiator_name)

        # Get volume
        vpsa_volume = self._get_vpsa_volume(volume)
        servers = self.vpsa._get_servers_attached_to_volume(vpsa_volume)
        attach = None
        for server in servers:
            if server == vpsa_srv:
                attach = server
                break
        # Attach volume to server
        if attach is None:
            self.vpsa_send_cmd('attach_volume',
                               vpsa_srv=vpsa_srv,
                               vpsa_vol=vpsa_volume['name'])

        data = self.vpsa_send_cmd('list_vol_attachments',
                                  vpsa_vol=vpsa_volume['name'])
        server = None
        servers = data.get('servers', [])
        for srv in servers:
            if srv['iqn'] == initiator_name:
                server = srv
                break

        if server is None:
            vol_name = (self._get_zadara_vol_template_name(
                        volume.name))
            raise zadara_exception.ZadaraAttachmentsNotFound(
                name=vol_name)

        target = server['target']
        lun = int(server['lun'])
        if None in [target, lun]:
            vol_name = (self._get_zadara_vol_template_name(
                        volume.name))
            raise zadara_exception.ZadaraInvalidAttachmentInfo(
                name=vol_name,
                reason=_('target=%(target)s, lun=%(lun)s') %
                {'target': target, 'lun': lun})

        ctrl_ip = self.vpsa._get_target_host(ctrl['ip'])
        properties = {'target_discovered': False,
                      'target_portal': (('%s:%s') % (ctrl_ip, '3260')),
                      'target_iqn': target,
                      'target_lun': lun,
                      'volume_id': volume.id,
                      'auth_method': 'CHAP',
                      'auth_username': ctrl['chap_user'],
                      'auth_password': ctrl['chap_passwd']}

        LOG.debug('Attach properties: %(properties)s',
                  {'properties': strutils.mask_password(properties)})
        return {'driver_volume_type':
                ('iser' if (self.configuration.safe_get('zadara_use_iser'))
                 else 'iscsi'), 'data': properties}

    def terminate_connection(self, volume, connector, **kwargs):
        """Detach volume from the initiator."""

        vpsa_volume = self._get_vpsa_volume(volume)

        if connector is None:
            # Detach volume from all servers
            # Get volume name
            self.vpsa._detach_vpsa_volume(vpsa_vol=vpsa_volume)
            return

        # Check if there are multiple attachments to the volume from the
        # same host. Terminate connection only for the last attachment from
        # the corresponding host.
        count = 0
        host = connector.get('host') if connector else None
        if host and volume.get('multiattach'):
            attach_list = volume.volume_attachment
            for attachment in attach_list:
                if (attachment['attach_status'] !=
                        fields.VolumeAttachStatus.ATTACHED):
                    continue
                if attachment.attached_host == host:
                    count += 1
        if count > 1:
            return

        # Get server name for IQN
        initiator_name = connector['initiator']

        vpsa_srv = self.vpsa._get_server_name(initiator_name, False)
        if not vpsa_srv:
            raise zadara_exception.ZadaraServerNotFound(name=initiator_name)

        if not vpsa_volume:
            raise cinder_exception.VolumeNotFound(volume_id=volume.id)

        # Detach volume from server
        self.vpsa._detach_vpsa_volume(vpsa_vol=vpsa_volume,
                                      vpsa_srv=vpsa_srv)

    def _update_volume_stats(self):
        """Retrieve stats info from volume group."""
        LOG.debug("Updating volume stats")
        backend_name = self.configuration.safe_get('volume_backend_name')
        storage_protocol = (constants.ISER if
                            (self.configuration.safe_get('zadara_use_iser'))
                            else constants.ISCSI)
        pool_name = self.configuration.zadara_vpsa_poolname
        (total, free, provisioned) = self.vpsa._get_pool_capacity(pool_name)
        data = dict(
            volume_backend_name=backend_name or self.__class__.__name__,
            vendor_name='Zadara Storage',
            driver_version=self.VERSION,
            storage_protocol=storage_protocol,
            reserved_percentage=self.configuration.reserved_percentage,
            QoS_support=False,
            multiattach=True,
            total_capacity_gb=total,
            free_capacity_gb=free
        )

        self._stats = data



#    Copyright (c) 2015 Intel Corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import fields

from cinder import objects
from cinder.objects import fields as c_fields
from cinder.tests.unit import fake_constants as fake


def fake_db_backup(**updates):
    db_backup = {
        'id': fake.BACKUP_ID,
        'user_id': fake.USER_ID,
        'project_id': fake.PROJECT_ID,
        'volume_id': fake.VOLUME_ID,
        'status': c_fields.BackupStatus.CREATING,
        'host': 'fake_host',
        'display_name': 'fake_name',
        'size': 5,
        'display_description': 'fake_description',
        'service_metadata': 'fake_metadata',
        'service': 'fake_service',
        'object_count': 5,
        'num_dependent_backups': 0,
        'backup_metadata': [
            {'key': 'key1', 'value': 'value1'},
            {'key': 'key2', 'value': 'value2'}
        ],
    }

    for name, field in objects.Backup.fields.items():
        if name in db_backup:
            continue
        if field.nullable:
            db_backup[name] = None
        elif field.default != fields.UnspecifiedDefault:
            db_backup[name] = field.default
        else:
            raise Exception('fake_db_backup needs help with %s' % name)

    if updates:
        db_backup.update(updates)

    return db_backup


def fake_backup_obj(context, **updates):
    return objects.Backup._from_db_object(context, objects.Backup(),
                                          fake_db_backup(**updates),
                                          expected_attrs=['metadata'])



# Copyright 2012 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinder.api import common


class ViewBuilder(common.ViewBuilder):

    def show(self, request, volume_type, brief=False):
        """Trim away extraneous volume type attributes."""
        trimmed = dict(id=volume_type.get('id'),
                       name=volume_type.get('name'),
                       is_public=volume_type.get('is_public'),
                       extra_specs=volume_type.get('extra_specs'),
                       description=volume_type.get('description'))
        return trimmed if brief else dict(volume_type=trimmed)

    def index(self, request, volume_types):
        """Index over trimmed volume types."""
        volume_types_list = [self.show(request, volume_type, True)
                             for volume_type in volume_types]
        return dict(volume_types=volume_types_list)
